"""Common utility functions"""
from typing import Type
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import List
from typing import Union

import types
import os
import inspect
import hashlib
import pickle
import functools
import torch
import numpy as np
import datamol as dm
import fsspec

from joblib import wrap_non_picklable_objects
from scipy.sparse import coo_matrix

from rdkit.Chem import rdMolDescriptors
from rdkit.Chem import rdMolAlign
from rdkit.Chem import SaltRemover
from molfeat.utils import datatype


FUNCTYPES = (types.FunctionType, types.MethodType, functools.partial)


def is_callable(func):
    r"""
    Check if func is a function or a callable
    """
    return func and (isinstance(func, FUNCTYPES) or callable(func))


def sha256sum(filepath: Union[str, os.PathLike]):
    """Return the sha256 sum hash of a file or a directory

    Args:
        filepath: The path to the file to compute the MD5 hash on.
    """
    if dm.fs.is_dir(filepath):
        files = list(dm.fs.glob(os.path.join(filepath, "**", "*")))
    else:
        files = [filepath]
    file_hash = hashlib.sha256()
    for filepath in sorted(files):
        with fsspec.open(filepath) as f:
            file_hash.update(f.read())  # type: ignore
    file_hash = file_hash.hexdigest()
    return file_hash


def get_class_name(cls: Type):
    """Get class full name

    Args:
        cls: name of the class
    """
    module = cls.__module__
    name = cls.__qualname__
    if module is not None and module != "__builtin__":
        name = module + "." + name
    return name


def _clean_mol_for_descriptors(
    mol, disconnect_metals: bool = False, remove_salt: bool = False, **kwargs
):
    """Clean molecule for descriptors computation
    Args:
        disconnect_metals: whether to disconnect or keep metal ions
        remove_salt: whether to remove salt or
    """
    mol = dm.to_mol(mol)
    mol = dm.standardize_mol(mol, disconnect_metals=disconnect_metals, **kwargs)
    if remove_salt:
        remover = SaltRemover.SaltRemover()
        mol = remover.StripMol(mol, dontRemoveEverything=True)
    return mol


def ensure_picklable(fn: Callable):
    """Ensure a function is picklable

    Args:
        fn: function to be pickled
    """
    if inspect.isfunction(fn) and fn.__name__ == "<lambda>":
        return wrap_non_picklable_objects(fn)
    return fn


def fn_to_hex(fn):
    """Pickle an object and return its hex representation

    Args:
        fn: object to pickle

    Returns:
        str: hex representation of object
    """
    bytes_str = pickle.dumps(ensure_picklable(fn))
    return bytes_str.hex()


def hex_to_fn(hex: str):
    """Load a hex string as a callable. Raise error on fail

    Args:
        hex: hex string to load as a callable

    Returns:
        callable: callable loaded from the hex string
    """
    # EN: pickling with pickle is probably faster
    fn = pickle.loads(bytes.fromhex(hex))
    return fn


def one_hot_encoding(
    val: int,
    allowable_set: Iterable,
    encode_unknown: bool = False,
    dtype: Callable = int,
):
    r"""Converts a single value to a one-hot vector.

    Args:
        val: class to be converted into a one hot vector
        allowable_set: a list or 1D array of allowed choices for val to take
        dtype: data type of the the return. Default = int.
        encode_unknown: whether to map inputs not in allowable set to an additional last element.

    Returns:
        A numpy 1D array of length len(allowable_set) + 1
    """

    encoding = np.zeros(len(allowable_set) + int(encode_unknown), dtype=dtype)
    # not using index of, in case, someone fuck up
    # and there are duplicates in the allowed choices
    for i, v in enumerate(allowable_set):
        if v == val:
            encoding[i] = 1
    if np.sum(encoding) == 0 and encode_unknown:  # aka not found
        encoding[-1] = 1
    return encoding


def _parse_to_evaluable_str(val: Callable):
    r"""
    Generates a string from an object, such that it can be evaluated.
    It only works with basic classes, or classes that implement a
    specific `__repr__()` method.

    Args:
        val: input object or function to repr

    Returns:
        val_str (str): String representation of the function
    """
    if inspect.isclass(val):
        if datatype.is_dtype_numpy(val):
            val_str = "np." + val.__name__
        elif datatype.is_dtype_tensor(val):
            val_str = str(val)
        else:
            val_str = val.__name__
    elif isinstance(val, str):
        val_str = '"{}"'.format(val)
    else:
        val_str = val.__repr__()
    return val_str


def filter_arguments(fn: Callable, params: dict):
    """Filter the argument of a function to only retain the valid ones

    Args:
        fn: Function for which arguments will be checked
        params: key-val dictionary of arguments to pass to the input function

    Returns:
        params_filtered (dict): dict of filtered arguments for the function
    """
    accepted_dict = inspect.signature(fn).parameters
    accepted_list = []
    for key in accepted_dict.keys():
        param = str(accepted_dict[key])
        if param[0] != "*":
            accepted_list.append(param)
    params_filtered = {key: params[key] for key in list(set(accepted_list) & set(params.keys()))}
    return params_filtered


def fold_count_fp(fp: Iterable, dim: int = 2**10, binary: bool = False):
    """Fast folding of a count fingerprint to the specified dimension

    Args:
        fp: iterable fingerprint
        dim: dimension of the folded array if not provided. Defaults to 2**10.
        binary: whether to fold into a binary array or take use a count vector

    Returns:
        folded: returns folded array to the provided dimension
    """
    if hasattr(fp, "GetNonzeroElements"):
        tmp = fp.GetNonzeroElements()
    elif hasattr(fp, "GetOnBits"):
        # try to get the dict of onbit
        on_bits = fp.GetOnBits()
        tmp = dict(zip(on_bits, np.ones(len(on_bits))))
    else:
        raise ValueError(f"Format {type(fp)} is not supported")
    out = (
        coo_matrix(
            (
                list(tmp.values()),
                (np.repeat(0, len(tmp)), [i % dim for i in tmp.keys()]),
            ),
            shape=(1, dim),
        )
        .toarray()
        .flatten()
    )
    if binary:
        out = np.clip(out, a_min=0, a_max=1)
    return out


def requires_conformer(calculator: Callable):
    """Decorator for any descriptor calculator that requires conformers"""

    # this is a method or __call__
    if inspect.getfullargspec(calculator).args[0] == "self":

        @functools.wraps(calculator)
        def calculator_wrapper(ref, mol, *args, **kwargs):
            mol = dm.to_mol(mol)
            if mol.GetNumConformers() < 1:
                raise ValueError("Expected a molecule with conformers information.")
            return calculator(ref, mol, *args, **kwargs)

    else:

        @functools.wraps(calculator)
        def calculator_wrapper(mol, *args, **kwargs):
            mol = dm.to_mol(mol)
            if mol.GetNumConformers() < 1:
                raise ValueError("Expected a molecule with conformers information.")
            return calculator(mol, *args, **kwargs)

    return calculator_wrapper


def requires_standardization(
    calculator: Callable = None,
    *,
    disconnect_metals: bool = True,
    remove_salt: bool = True,
    **standardize_kwargs,
):
    """Decorator for any descriptor calculator that required standardization of the molecules
    Args:
        calculator: calculator to wrap
        disconnect_metals: whether to force metal disconnection
        remove_salt: whether to remove salt from the molecule
    """

    def _standardize_mol(calculator):
        @functools.wraps(calculator)
        def wrapped_function(mol, *args, **kwargs):
            mol = _clean_mol_for_descriptors(
                mol,
                disconnect_metals=disconnect_metals,
                remove_salt=remove_salt,
                **standardize_kwargs,
            )
            return calculator(mol, *args, **kwargs)

        @functools.wraps(calculator)
        def class_wrapped_function(ref, mol, *args, **kwargs):
            if not getattr(ref, "do_not_standardize", False):
                mol = _clean_mol_for_descriptors(
                    mol,
                    disconnect_metals=disconnect_metals,
                    remove_salt=remove_salt,
                    **standardize_kwargs,
                )
            return calculator(ref, mol, *args, **kwargs)

        if inspect.getfullargspec(calculator).args[0] == "self":
            return class_wrapped_function
        return wrapped_function

    if calculator is not None:
        return _standardize_mol(calculator)
    return _standardize_mol


def concat_dict(prop_dict: dict, new_name: str, order: Optional[Iterable[str]] = None):
    """Concat properties in dict into a single key dict

    Args:
        prop_dict (dict): Input dict of property names and their computed values
        new_name (str): new name under which the concatenated property dict will be returned
        order: Optional list of key that specifies the order in which concatenation should be done. Sorting list by default

    Returns:
        dict: dictionary of concatenated output values with a single key corresponding to new_name
    """
    if not order:
        order = list(sorted(prop_dict.keys()))

    if len(order) > 0:
        concatenated_val = np.concatenate([prop_dict[x] for x in order], axis=1)
        output_dict = {new_name: concatenated_val}
    return output_dict


def pack_graph(
    batch_G: List[torch.FloatTensor],
    batch_x: List[torch.FloatTensor],
):
    r"""
    Pack a batch of graph and atom features into a single graph

    Args:
        batch_G: List of adjacency graph, each of size (n_i, n_i).
        batch_x: List of atom feature matrices, each of size (n_i, F), F being the number of features

    Returns:
        new_batch_G, new_batch_x: torch.LongTensor 2D, torch.Tensor 2D
            This tuple represents a new arbitrary graph that contains the whole batch,
            and the corresponding atom feature matrix. new_batch_G has a size (N, N), with :math:`N = \sum_i n_i`,
            while new_batch_x has size (N,D)
    """

    new_batch_x = torch.cat(tuple(batch_x), dim=0)
    n_neigb = new_batch_x.shape[0]
    # should be on the same device
    new_batch_G = batch_G[0].new_zeros((n_neigb, n_neigb))
    cur_ind = 0
    for g in batch_G:
        g_size = g.shape[0] + cur_ind
        new_batch_G[cur_ind:g_size, cur_ind:g_size] = g
        cur_ind = g_size
    return new_batch_G, new_batch_x


def pack_bits(obj, protocol=4):
    """Pack an object into a bits representation

    Args:
        obj: object to pack

    Returns:
        bytes: byte-packed version of object
    """
    return pickle.dumps(obj, protocol=protocol)


def unpack_bits(bvalues):
    """Pack an object into a bits representation

    Args:
        bvalues: bytes to be unpacked

    Returns:
        obj: object that was packed
    """
    return pickle.loads(bvalues)


def align_conformers(
    mols: List[dm.Mol],
    ref_id: int = 0,
    copy: bool = True,
    conformer_id: int = -1,
):
    """Align a list of molecules to a reference molecule.

    Note: consider adding me to `datamol`.

    Args:
        mols: List of molecules to align. All the molecules must have a conformer.
        ref_id: Index of the reference molecule. By default, the first molecule in the list
            will be used as reference.
        copy: Whether to copy the molecules before performing the alignement.
        conformer_id: Conformer id to use.

    Returns:
        mols: The aligned molecules.
        scores: The score of the alignement.
    """

    # Check all input molecules has a conformer
    if not all([mol.GetNumConformers() >= 1 for mol in mols]):
        raise ValueError("One or more input molecules is missing a conformer.")

    # Make a copy of the molecules since they are going to be modified
    if copy:
        mols = [dm.copy_mol(mol) for mol in mols]

    # Compute Crippen contributions for every atoms and molecules
    crippen_contribs = [rdMolDescriptors._CalcCrippenContribs(mol) for mol in mols]

    # Split reference and probe molecules
    crippen_contrib_ref = crippen_contribs[ref_id]
    crippen_contrib_probes = crippen_contribs
    mol_ref = mols[ref_id]
    mol_probes = mols

    # Loop and align
    scores = []
    for i, mol in enumerate(mol_probes):
        crippenO3A = rdMolAlign.GetCrippenO3A(
            prbMol=mol,
            refMol=mol_ref,
            prbCrippenContribs=crippen_contrib_probes[i],
            refCrippenContribs=crippen_contrib_ref,
            prbCid=conformer_id,
            refCid=conformer_id,
            maxIters=50,
        )
        crippenO3A.Align()

        scores.append(crippenO3A.Score())

    scores = np.array(scores)

    return mols, scores



from molfeat.trans.graph.adj import AdjGraphTransformer
from molfeat.trans.graph.adj import TopoDistGraphTransformer
from molfeat.trans.graph.adj import DistGraphTransformer3D
from molfeat.trans.graph.adj import CompleteGraphTransformer
from molfeat.trans.graph.adj import DGLGraphTransformer
from molfeat.trans.graph.adj import GraphTransformer
from molfeat.trans.graph.adj import PYGGraphTransformer
from molfeat.trans.graph.tree import MolTreeDecompositionTransformer



from typing import Union

import datamol as dm
from loguru import logger
from molfeat.utils import requires

if requires.check("map4"):
    from map4 import MAP4Calculator
else:
    MAP4Calculator = requires.mock("map4")


def MAP4(
    x: Union[dm.Mol, str],
    dimensions: int = 2048,
    radius: int = 2,
    is_counted: bool = False,
    is_folded: bool = True,
    return_strings: bool = False,
    **kwargs,
):
    """Compute MHFP fingerprint

    Args:
        x: input molecule
        dimensions (int, optional): Length of the fingerprint (default: 2048).
        radius (int, optional): Radius of the fingerprint (default: 3)
        is_counted (bool, optional): Whether to use counted fingerprints (default: True)
        is_folded (bool, optional): Whether to fold the fingerprint (default: True)
        return_strings (bool, optional): Whether to return strings values (default: False)

    Returns:
        fp: fingerprint
    """
    if not requires.check("map4"):
        logger.error(
            "`map4` is not available, please install it: https://github.com/reymond-group/map4"
        )
        raise ImportError("Cannot import `map4`")

    if isinstance(x, str):
        x = dm.to_mol(x)
    map4_encoder = MAP4Calculator(
        dimensions=dimensions,
        radius=radius,
        is_counted=is_counted,
        is_folded=is_folded,
        return_strings=return_strings,
    )
    encoded_fp = map4_encoder.calculate(x)
    return encoded_fp



from typing import Union
from typing import Callable
from typing import Optional
from typing import Dict
from typing import Iterable

from collections import defaultdict

import importlib
import inspect
import datamol as dm
import numpy as np

from molfeat._version import __version__ as MOLFEAT_VERSION
from molfeat.calc.base import SerializableCalculator
from molfeat.calc._atom_bond_features import bond_type_one_hot
from molfeat.calc._atom_bond_features import bond_is_conjugated
from molfeat.calc._atom_bond_features import bond_is_in_ring
from molfeat.calc._atom_bond_features import bond_direction_one_hot
from molfeat.calc._atom_bond_features import bond_stereo_one_hot
from molfeat.calc._atom_bond_features import pairwise_ring_membership
from molfeat.calc._atom_bond_features import pairwise_2D_dist
from molfeat.calc._atom_bond_features import pairwise_bond_indicator
from molfeat.calc._atom_bond_features import pairwise_dist_indicator
from molfeat.utils import datatype
from molfeat.utils.commons import concat_dict
from molfeat.utils.commons import hex_to_fn
from molfeat.utils.commons import fn_to_hex


class BondCalculator(SerializableCalculator):
    """
    A class for bond featurizer which loops over all bonds in a molecule and
    featurizes them with the ``featurizer_funcs``. The constructed graph is assumed to be
    a bi-directed graph by default.
    """

    DEFAULT_FEATURIZER = {
        "bond_type_one_hot": bond_type_one_hot,
        "bond_stereo_one_hot": bond_stereo_one_hot,
        "bond_is_in_ring": bond_is_in_ring,
        "bond_is_conjugated": bond_is_conjugated,
        "bond_direction_one_hot": bond_direction_one_hot,
    }

    def __init__(
        self,
        featurizer_funcs: Union[list, dict] = None,
        self_loop: bool = False,
        concat: bool = True,
        name: str = "he",
    ):
        """
        Init function of the bond property calculator

        Args:
            featurizer_funcs: Mapping feature name to the featurization function.
            self_loop: Whether self loops will be added. Default to False. If True, an additional
                column of binary values to indicate the identity of self loops will be added.
                The other features of the self loops will be zero.
            concat: Whether to concat all the data into a single value in the output dict
            name: Name of the key name of the concatenated features
        """
        self._input_kwargs = locals().copy()
        self._input_kwargs.pop("self")
        # remove featurizer_funcs too
        self._input_kwargs.pop("featurizer_funcs", None)
        self._toy_mol = dm.to_mol("CO")
        self._feat_sizes = dict()
        if featurizer_funcs is None:
            featurizer_funcs = self.DEFAULT_FEATURIZER
        if not isinstance(featurizer_funcs, dict):
            get_name = lambda x: getattr(x, "__name__", repr(x))
            featurizer_funcs = dict((get_name(x), x) for x in featurizer_funcs)
        self.featurizer_funcs = featurizer_funcs
        self._self_loop = self_loop
        self.concat = concat
        self.name = name
        for k in self.featurizer_funcs.keys():
            self.feat_size(feat_name=k)
        if self._self_loop:
            self._feat_sizes["self_loop"] = 1

    def to_state_dict(self):
        """Convert the Atom calculator to a state dict
        Due to some constraints and cross-version compatibility,  the featurizer functions
        need to be pickled and not just list
        """
        state_dict = {}
        state_dict["name"] = self.__class__.__name__
        state_dict["module"] = self.__class__.__module__
        state_dict["args"] = self._input_kwargs

        featurizer_fn_pickled = {}
        for fname, ffunc in self.featurizer_funcs.items():
            featurizer_fn_pickled[fname] = fn_to_hex(ffunc)
        state_dict["args"]["featurizer_funcs"] = featurizer_fn_pickled
        state_dict["_molfeat_version"] = MOLFEAT_VERSION
        signature = inspect.signature(self.__init__)
        val = {
            k: v.default
            for k, v in signature.parameters.items()
            #    if v.default is not inspect.Parameter.empty
        }
        to_remove = [k for k in state_dict["args"] if k not in val.keys()]
        for k in to_remove:
            state_dict["args"].pop(k)
        return state_dict

    @classmethod
    def from_state_dict(cls, state_dict, override_args: Optional[dict] = None):
        """Create an instance of an atom calculator from a state dict

        Args:
            state_dict: state dictionary to use to create the atom calculator
            override_args: optional dictionary of arguments to override the ones in the state dict
                at construction of the new object
        """
        # EN: at this moment, version compatibility is not enforced
        cls_name = state_dict.get("name", cls.__name__)
        module_name = state_dict.get("module", cls.__module__)
        module = importlib.import_module(module_name)
        klass = getattr(module, cls_name)

        kwargs = state_dict["args"].copy()
        # now we need to unpickle the featurizer functions
        featurizer_fn_pickled = kwargs.pop("featurizer_funcs", None)
        if featurizer_fn_pickled is not None:
            featurizer_fn_loaded = {}
            for k, v in featurizer_fn_pickled.items():
                featurizer_fn_loaded[k] = hex_to_fn(v)
            kwargs["featurizer_funcs"] = featurizer_fn_loaded
        kwargs.update(**(override_args or {}))
        return klass(**kwargs)

    def _concat(self, data_dict: Dict[str, Iterable]):
        """Concatenate the data into a single value

        Args:
            data_dict: mapping of feature names to tensor/arrays
        Returns:
            concatenated_dict: a dict with a single key where all array have been concatenated
        """
        return concat_dict(data_dict, new_name=self.name)

    def feat_size(self, feat_name: Optional[str] = None):
        """Get the feature size for ``feat_name``.

        When there is only one feature, ``feat_name`` can be None.

        Args:
            feat_name: Feature for query.

        Returns:
            int: Feature size for the feature with name ``feat_name``. Default to None.
        """
        if feat_name is None:
            assert (
                len(self.featurizer_funcs) == 1
            ), "feat_name should be provided if there are more than one features"
            feat_name = list(self.featurizer_funcs.keys())[0]

        if feat_name not in self.featurizer_funcs:
            raise ValueError(
                "Expect feat_name to be in {}, got {}".format(
                    list(self.featurizer_funcs.keys()), feat_name
                )
            )
        if feat_name not in self._feat_sizes:
            bond = self._toy_mol.GetBondWithIdx(0)
            self._feat_sizes[feat_name] = len(self.featurizer_funcs[feat_name](bond))
        return self._feat_sizes[feat_name]

    def __len__(self):
        """Get length of the property estimator"""
        return sum(v for k, v in self._feat_sizes.items() if k != self.name)

    def __call__(self, mol: Union[dm.Mol, str], dtype: Callable = None, **kwargs):
        """Featurize all bonds in a molecule.

        Args:
            mol: the molecule of interest
            dtype: requested data type

        Returns:
            dict: For each function in self.featurizer_funcs with the key ``k``,
                store the computed feature under the key ``k``.
        """
        mol = dm.to_mol(mol)
        num_bonds = mol.GetNumBonds()
        bond_features = defaultdict(list)

        # Compute features for each bond
        for i in range(num_bonds):
            bond = mol.GetBondWithIdx(i)
            for feat_name, feat_func in self.featurizer_funcs.items():
                feat = feat_func(bond)
                bond_features[feat_name].extend([feat, feat.copy()])

        # Stack the features and convert them to float arrays
        processed_features = dict()
        for feat_name, feat_list in bond_features.items():
            feat = np.stack(feat_list)
            processed_features[feat_name] = feat

        if self._self_loop and num_bonds > 0:
            num_atoms = mol.GetNumAtoms()
            for feat_name in processed_features:
                feats = processed_features[feat_name]
                # add a new label that says the feat are not self loop
                # feats = np.concatenate([feats, np.zeros((feats.shape[0], 1))], axis=1)
                # add a label at the last position that says it's a selfloop
                add_edges = np.zeros((num_atoms, feats.shape[1]))
                # self_loop_feats[:, -1] = 1
                feats = np.concatenate([feats, add_edges], axis=0)
                processed_features[feat_name] = feats
            self_loop_feats = np.concatenate(
                [np.zeros((num_bonds * 2, 1)), np.ones((num_atoms, 1))]
            )

            processed_features["self_loop"] = self_loop_feats

        if self._self_loop and num_bonds == 0:
            num_atoms = mol.GetNumAtoms()
            old_concat = self.concat
            self.concat = False
            processed_features = self(self._toy_mol)
            self.concat = old_concat
            for feat_name in processed_features:
                feats = processed_features[feat_name]
                feats = np.zeros((num_atoms, feats.shape[1]))
                processed_features[feat_name] = feats
        if self.concat and (num_bonds > 0 or self._self_loop):
            processed_features = self._concat(processed_features)
        if dtype is not None:
            for feat_name, feat in processed_features.items():
                feat = datatype.cast(feat, dtype=dtype)
                processed_features[feat_name] = feat

        return processed_features


class EdgeMatCalculator(BondCalculator):
    """Generate edge featurizer matrix"""

    DEFAULT_PAIRWISE_FEATURIZER = {
        "pairwise_2D_dist": pairwise_2D_dist,
        # "pairwise_3D_dist": pairwise_3D_dist,
        "pairwise_ring_membership": pairwise_ring_membership,
    }

    def __init__(
        self,
        featurizer_funcs: Union[list, dict] = None,
        pairwise_atom_funcs: Union[list, dict, str] = "default",
        name: str = "he",
    ):
        """
        Init function of the edge matrix property calculator

        Args:
            featurizer_funcs: Mapping feature name to the featurization function.
            pairwise_atom_funcs: Mapping feature name to pairwise featurization function.
                Use the keywords "default" for the default values
        """
        if pairwise_atom_funcs == "default":
            pairwise_atom_funcs = self.DEFAULT_PAIRWISE_FEATURIZER
        if not isinstance(pairwise_atom_funcs, dict):
            get_name = lambda x: getattr(x, "__name__", repr(x))
            pairwise_atom_funcs = dict((get_name(x), x) for x in pairwise_atom_funcs)
        self.pairwise_atom_funcs = pairwise_atom_funcs
        super().__init__(featurizer_funcs=featurizer_funcs, concat=True, name=name)
        # add conf data to toy mol
        self._toy_mol = dm.conformers.generate(self._toy_mol, n_confs=1, minimize_energy=False)
        for k in self.pairwise_atom_funcs.keys():
            self.feat_size(feat_name=k)

    def to_state_dict(self):
        """Convert the Atom calculator to a state dict
        Due to some constraints and cross-version compatibility,  the featurizer functions
        need to be pickled and not just list
        """
        state_dict = super().to_state_dict()
        # repeat for the pairwise one
        pairwise_atom_fn_pickled = {}
        for fname, ffunc in self.pairwise_atom_funcs.items():
            pairwise_atom_fn_pickled[fname] = fn_to_hex(ffunc)
        state_dict["args"]["pairwise_atom_funcs"] = pairwise_atom_fn_pickled
        return state_dict

    @classmethod
    def from_state_dict(cls, state_dict, override_args: Optional[dict] = None):
        """Create an instance of an atom calculator from a state dict

        Args:
            state_dict: state dictionary to use to create the atom calculator
            override_args: optional dictionary of arguments to override the ones in the state dict
                at construction of the new object
        """
        # EN: at this moment, version compatibility is not enforced
        cls_name = state_dict.get("name", cls.__name__)
        module_name = state_dict.get("module", cls.__module__)
        module = importlib.import_module(module_name)
        klass = getattr(module, cls_name)

        kwargs = state_dict["args"].copy()
        # now we need to unpickle the featurizer functions
        featurizer_fn_pickled = kwargs.pop("featurizer_funcs", None)
        if featurizer_fn_pickled is not None:
            featurizer_fn_loaded = {}
            for k, v in featurizer_fn_pickled.items():
                featurizer_fn_loaded[k] = hex_to_fn(v)
            kwargs["featurizer_funcs"] = featurizer_fn_loaded

        pairwise_atom_fn_pickled = kwargs.pop("pairwise_atom_funcs", None)
        if pairwise_atom_fn_pickled is not None:
            pairwise_atom_fn_loaded = {}
            for k, v in pairwise_atom_fn_pickled.items():
                pairwise_atom_fn_loaded[k] = hex_to_fn(v)
            kwargs["pairwise_atom_funcs"] = pairwise_atom_fn_loaded
        kwargs.update(**(override_args or {}))
        return klass(**kwargs)

    def feat_size(self, feat_name: Optional[str] = None):
        """Get the feature size for ``feat_name``.

        Args:
            feat_name: Feature for query.

        Returns:
            int: Feature size for the feature with name ``feat_name``. Default to None.
        """
        if feat_name not in self.featurizer_funcs and feat_name not in self.pairwise_atom_funcs:
            raise ValueError(
                "Expect feat_name to be in {}, got {}".format(
                    list(self.featurizer_funcs.keys()), feat_name
                )
            )
        if feat_name not in self._feat_sizes:
            if feat_name in self.featurizer_funcs:
                bond = self._toy_mol.GetBondWithIdx(0)
                self._feat_sizes[feat_name] = len(self.featurizer_funcs[feat_name](bond))
            elif feat_name in self.pairwise_atom_funcs:
                self._feat_sizes[feat_name] = self.pairwise_atom_funcs[feat_name](
                    self._toy_mol
                ).shape[-1]
            else:
                raise ValueError(f"Feature name {feat_name} is not defined !")
        return self._feat_sizes[feat_name]

    def __call__(self, mol: Union[dm.Mol, str], dtype: Callable = None, flat: bool = True):
        """Featurize all bonds in a molecule.

        Args:
            mol: the molecule of interest
            dtype: requested data type
            flat: whether to return a collapsed N^2, M or a N, N, M matrix

        Returns:
            dict: For each function in self.featurizer_funcs with the key ``k``,
                store the computed feature under the key ``k``.
        """

        mol = dm.to_mol(mol)
        num_bonds = mol.GetNumBonds()
        num_atoms = mol.GetNumAtoms()
        feat_size = len(self)
        edge_matrix = None

        if self.pairwise_atom_funcs is not None:
            feat_size -= sum(self._feat_sizes[x] for x in self.pairwise_atom_funcs.keys())
        if self.featurizer_funcs is not None and len(self.featurizer_funcs) > 0:
            edge_matrix = np.zeros((num_atoms, num_atoms, feat_size))
            # Compute features for each bond
            for i in range(num_bonds):
                bond = mol.GetBondWithIdx(i)
                a_idx_1 = bond.GetBeginAtomIdx()
                a_idx_2 = bond.GetEndAtomIdx()
                bond_features = defaultdict(list)
                for feat_name, feat_func in self.featurizer_funcs.items():
                    feat = feat_func(bond)
                    bond_features[feat_name].extend([feat])
                bond_features = self._concat(bond_features)[self.name]
                edge_matrix[a_idx_1, a_idx_2] = bond_features
                edge_matrix[a_idx_2, a_idx_1] = bond_features

            edge_matrix = edge_matrix.reshape(-1, feat_size)
        if self.pairwise_atom_funcs is not None:
            pwise_features = dict()
            for pname, pfunc in self.pairwise_atom_funcs.items():
                pwise_features[pname] = pfunc(mol)
            pwise_features = self._concat(pwise_features)[self.name]
            if edge_matrix is not None:
                edge_matrix = np.concatenate([edge_matrix, pwise_features], axis=-1)
            else:
                edge_matrix = pwise_features
        if not flat:
            edge_matrix = edge_matrix.reshape(num_atoms, num_atoms, -1)
        if dtype is not None:
            edge_matrix = datatype.cast(edge_matrix, dtype=dtype)
        return {self.name: edge_matrix}


class DGLCanonicalBondCalculator(BondCalculator):
    DEFAULT_FEATURIZER = {
        "bond_type_one_hot": bond_type_one_hot,
        "bond_is_conjugated": bond_is_conjugated,
        "bond_is_in_ring": bond_is_in_ring,
        "bond_stereo_one_hot": bond_stereo_one_hot,
    }

    def _concat(self, data_dict: Dict[str, Iterable]):
        """Concatenate the data into a single value

        Args:
            data_dict: mapping of feature names to tensor/arrays
        Returns:
            concatenated_dict: a dict with a single key where all array have been concatenated
        """
        return concat_dict(data_dict, new_name=self.name, order=list(self.featurizer_funcs.keys()))


class DGLWeaveEdgeCalculator(EdgeMatCalculator):
    """Edge featurizer used by WeaveNets

    The edge featurization is introduced in `Molecular Graph Convolutions:
    Moving Beyond Fingerprints <https://arxiv.org/abs/1603.00856>`__.

    This featurization is performed for a complete graph of atoms with self loops added,
    which considers the following default:

    * Number of bonds between each pairs of atoms
    * One-hot encoding of bond type if a bond exists between a pair of atoms
    * Whether a pair of atoms belongs to a same ring

    """

    DEFAULT_FEATURIZER = {}
    DEFAULT_PAIRWISE_FEATURIZER = {
        "pairwise_dist_indicator": pairwise_dist_indicator,
        "pairwise_bond_indicator": pairwise_bond_indicator,
        "pairwise_ring_membership": pairwise_ring_membership,
    }

    def _concat(self, data_dict: Dict[str, Iterable]):
        """Concatenate the data into a single value

        Args:
            data_dict: mapping of feature names to tensor/arrays
        Returns:
            concatenated_dict: a dict with a single key where all array have been concatenated
        """

        # To reproduce DGLDefault, we need to keep the order of dict insertion
        return concat_dict(
            data_dict, new_name=self.name, order=list(self.pairwise_atom_funcs.keys())
        )



from datetime import datetime
from typing import List, Literal, Optional, Union

import datamol as dm
from pydantic import BaseModel, ConfigDict, Field


def get_model_init(card):
    """Get the model initialization code

    Args:
        card: model card to use
    """
    if card.group == "all" and card.type != "pretrained":
        import_statement = "from molfeat.trans import MoleculeTransformer"
        loader_statement = f"MoleculeTransformer(featurizer='{card.name}', dtype=float)"
    elif card.group in ["rdkit", "fp", "shape"]:
        import_statement = "from molfeat.trans.fp import FPVecTransformer"
        loader_statement = f"FPVecTransformer(kind='{card.name}', dtype=float)"
    elif card.group == "dgllife":
        import_statement = "from molfeat.trans.pretrained import PretrainedDGLTransformer"
        loader_statement = f"PretrainedDGLTransformer(kind='{card.name}', dtype=float)"
    elif card.group == "graphormer":
        import_statement = "from molfeat.trans.pretrained import GraphormerTransformer"
        loader_statement = f"GraphormerTransformer(kind='{card.name}', dtype=float)"
    elif card.group == "fcd":
        import_statement = "from molfeat.trans.pretrained import FCDTransformer"
        loader_statement = "FCDTransformer()"
    elif card.group == "pharmacophore":
        name = card.name.split("-")[-1]
        if card.require_3D:
            import_class = "Pharmacophore3D"
        else:
            import_class = "Pharmacophore2D"
        import_statement = f"from molfeat.trans.base import MoleculeTransformer\nfrom molfeat.calc.pharmacophore import {import_class}"
        loader_statement = (
            f"MoleculeTransformer(featurizer={import_class}(factory='{name}'), dtype=float)"
        )
    elif card.group == "huggingface":
        import_statement = (
            "from molfeat.trans.pretrained.hf_transformers import PretrainedHFTransformer"
        )
        loader_statement = (
            f"PretrainedHFTransformer(kind='{card.name}', notation='{card.inputs}', dtype=float)"
        )
    else:
        raise ValueError(f"Unknown model group {card.group}")
    return import_statement, loader_statement


class ModelInfo(BaseModel):
    model_config = ConfigDict(
        protected_namespaces=(
            "protected_",
        )  # Prevents warning from usage of model_ prefix in fields
    )

    name: str
    inputs: str = "smiles"
    type: Literal["pretrained", "hand-crafted", "hashed", "count"]
    version: int = 0
    group: Optional[str] = "all"
    submitter: str
    description: str
    representation: Literal["graph", "line-notation", "vector", "tensor", "other"]
    require_3D: Optional[bool] = False
    tags: Optional[List[str]] = []
    authors: Optional[List[str]]
    reference: Optional[str] = None
    created_at: datetime = Field(default_factory=datetime.now)
    sha256sum: Optional[str] = None
    model_usage: Optional[str] = None

    def path(self, root_path: str):
        """Generate the folder path where to save this model

        Args:
            root_path: path to the root folder
        """
        version = str(self.version or 0)
        return dm.fs.join(root_path, self.group, self.name, version)

    def match(self, new_card: Union["ModelInfo", dict], match_only: Optional[List[str]] = None):
        """Compare two model card information and returns True if they are the same

        Args:
            new_card: card to search for in the modelstore
            match_only: list of minimum attribute that should match between the two model information
        """

        self_content = self.model_dump().copy()
        if not isinstance(new_card, dict):
            new_card = new_card.model_dump()
        new_content = new_card.copy()
        # we always remove the datetime field
        self_content.pop("created_at", None)
        new_content.pop("created_at", None)
        if match_only is not None:
            self_content = {k: self_content.get(k) for k in match_only}
            new_content = {k: new_content.get(k) for k in match_only}
        return self_content == new_content

    def set_usage(self, usage: str):
        """Set the usage of the model

        Args:
            usage: usage of the model
        """
        self.model_usage = usage

    def usage(self):
        """Return the usage of the model"""
        if self.model_usage is not None and self.model_usage:
            return self.model_usage
        import_statement, loader_statement = get_model_init(self)
        comment = "# sanitize and standardize your molecules if needed"
        if self.require_3D:
            comment += "\n# <generate 3D coordinates here> "
        usage = f"""
        import datamol as dm
        {import_statement}
        smiles = dm.freesolv().iloc[:100].smiles
        {comment}
        transformer = {loader_statement}
        features = transformer(smiles)
        """
        usage = "\n".join([x.strip() for x in usage.split("\n")])
        return usage



import unittest as ut
import datamol as dm
import dgl
import torch
import pytest
from molfeat.trans.graph.adj import AdjGraphTransformer
from molfeat.trans.graph.adj import DGLGraphTransformer
from molfeat.trans.graph.tree import MolTreeDecompositionTransformer
from molfeat.calc.tree import TreeDecomposer
from molfeat.utils import requires


class TestMolTreeDecomposition(ut.TestCase):
    r"""Test cases for Tree decomposition"""
    smiles = [
        "CCOc1c(OC)cc(CCN)cc1OC",
        "COc1cc(CCN)cc(OC)c1OC",
        "C[C@@H]([NH3+])Cc1c2ccoc2c(Br)c2ccoc12",
    ]
    mols = [dm.to_mol(smile) for smile in smiles]
    decomposer = TreeDecomposer(cache=False)

    def test_clusters(self):
        sm = "C1=CC=CC=C1"
        mol = dm.to_mol(sm)
        nodes, edges, frags = self.decomposer(mol)
        expected_nodes = [list(range(6))]
        self.assertEqual(len(nodes), 1)
        self.assertEqual(len(edges), 0)
        self.assertSetEqual(set(nodes[0]), set(expected_nodes[0]))

        sm2 = "C12=CC=CC=C1C(=C(S2)C(=O)NC)Cl"
        mol2 = dm.to_mol(sm2)
        nodes, edges, frags = self.decomposer(mol2)
        self.assertEqual(len(nodes), 8)
        self.assertEqual(len(edges), 7)

        expected_frags = ["cC", "C", "cCl", "CN", "c1ccsc1", "C=O", "c1ccccc1"]
        self.assertSetEqual(set(frags), set(expected_frags))

    def test_moltree_transformer(self):
        transf = MolTreeDecompositionTransformer()
        transf.fit(self.mols)
        tree, ids = transf(self.smiles, ignore_errors=True)
        self.assertIsInstance(tree, list)
        self.assertTrue(isinstance(tree[0], dgl.DGLGraph))


@pytest.mark.xfail(not requires.check("dgllife"), reason="3rd party module dgllife is missing")
class TestGraphTransformer(ut.TestCase):
    r"""Test cases for AdjGraphTransformer"""
    smiles = [
        "CCOc1c(OC)cc(CCN)cc1OC",
        "COc1cc(CCN)cc(OC)c1OC",
        "C[C@@H]([NH3+])Cc1c2ccoc2c(Br)c2ccoc12",
        "XXXRRR",
    ]
    mols = [dm.to_mol(x) for x in smiles]

    def test_adj_transformer(self):
        transf = AdjGraphTransformer(self_loop=True, dtype=torch.float)
        data, ids = transf(self.smiles, enforce_dtype=True, ignore_errors=True)
        # graphs and node features tuple
        self.assertEqual(len(data[0]), 2)
        self.assertEqual(ids, [0, 1, 2])
        (graphs, node_feats) = zip(*data)
        self.assertTrue(torch.is_tensor(graphs[0]))
        # get graph and node feat
        mat_sum = self.mols[0].GetNumBonds() * 2 + self.mols[0].GetNumAtoms()
        self.assertAlmostEqual(graphs[0].sum().item(), mat_sum)

    def test_dgl_transformer(self):
        transf = DGLGraphTransformer()
        graphs, ids = transf(self.smiles, ignore_errors=True)
        self.assertEqual(ids, [0, 1, 2])
        self.assertTrue(isinstance(graphs[0], dgl.DGLGraph))
        self.assertTrue(graphs[0].number_of_nodes(), self.mols[0].GetNumAtoms())

        with self.assertRaises(ValueError) as context:
            _ = transf(self.smiles, ignore_errors=False)
            self.assertTrue("transform molecule at index 3" in str(context.exception))



from typing import Optional
from typing import List
from typing import Any


import torch
import numpy as np
import datamol as dm

from rdkit.Chem import rdchem
from rdkit.Chem import AllChem
from rdkit.Chem import GetSymmSSSR
from rdkit.Chem.rdmolops import GetDistanceMatrix
from rdkit.Chem.rdmolops import Get3DDistanceMatrix
from molfeat.utils.commons import requires_conformer
from molfeat.data import get_df as get_data
from molfeat.utils.commons import one_hot_encoding


ATOM_LIST = [
    "C",
    "N",
    "O",
    "S",
    "F",
    "Si",
    "P",
    "Cl",
    "Br",
    "Mg",
    "Na",
    "Ca",
    "Fe",
    "As",
    "Al",
    "I",
    "B",
    "V",
    "K",
    "Tl",
    "Yb",
    "Sb",
    "Sn",
    "Ag",
    "Pd",
    "Co",
    "Se",
    "Ti",
    "Zn",
    "H",
    "Li",
    "Ge",
    "Cu",
    "Au",
    "Ni",
    "Cd",
    "In",
    "Mn",
    "Zr",
    "Cr",
    "Pt",
    "Hg",
    "Pb",
]
DGLLIFE_WEAVE_ATOMS = ["H", "C", "N", "O", "F", "P", "S", "Cl", "Br", "I"]

ATOM_DEGREE_LIST = list(range(11))
ATOM_TOTAL_DEGREE_LIST = list(range(6))
EXPLICIT_VALENCE_LIST = list(range(1, 7))
IMPLICIT_VALENCE_LIST = list(range(7))
# EN: (see molfeat/issues/141) changed from enumeration from rdkit.Hybridization
# to stating the list of hybridation because of RDKit introducing
# SP2D in  '2022.09.1'
HYBRIDIZATION_LIST = [
    rdchem.HybridizationType.UNSPECIFIED,
    rdchem.HybridizationType.SP3D2,
    rdchem.HybridizationType.SP3D,
    rdchem.HybridizationType.SP3,
    rdchem.HybridizationType.SP2,
    rdchem.HybridizationType.SP,
    rdchem.HybridizationType.S,
]


DGLLIFE_HYBRIDIZATION_LIST = [
    rdchem.HybridizationType.SP,
    rdchem.HybridizationType.SP2,
    rdchem.HybridizationType.SP3,
    rdchem.HybridizationType.SP3D,
    rdchem.HybridizationType.SP3D2,
]
ATOM_NUM_H_LIST = [0, 1, 2, 3, 4]
CHARGE_LIST = [-3, -2, -1, 0, 1, 2, 3]
RADICAL_ELECTRON_LIST = list(range(5))
CHIRAL_TYPES = [
    rdchem.ChiralType.CHI_UNSPECIFIED,
    rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
    rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
    rdchem.ChiralType.CHI_OTHER,
]

DGLLIFE_WEAVE_CHIRAL_TYPES = [
    rdchem.ChiralType.CHI_TETRAHEDRAL_CW,
    rdchem.ChiralType.CHI_TETRAHEDRAL_CCW,
]

BOND_TYPES = [
    rdchem.BondType.SINGLE,
    rdchem.BondType.DOUBLE,
    rdchem.BondType.TRIPLE,
    rdchem.BondType.AROMATIC,
]


BOND_STEREO = [
    rdchem.BondStereo.STEREONONE,
    rdchem.BondStereo.STEREOANY,
    rdchem.BondStereo.STEREOZ,
    rdchem.BondStereo.STEREOE,
    rdchem.BondStereo.STEREOCIS,
    rdchem.BondStereo.STEREOTRANS,
]

BOND_DIRECTION = [
    rdchem.BondDir.NONE,
    rdchem.BondDir.ENDUPRIGHT,
    rdchem.BondDir.ENDDOWNRIGHT,
]


def atom_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
    number: bool = False,
):
    """One hot encoding for the type of an atom.

    Args:
        atom : RDKit atom instance.
        allowable_set : Atom types to consider. Default is the full periodic table
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)
        number: Whether to use the atomic number of element symbol

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = ATOM_LIST if not number else list(range(1, 101))
    atom_val = atom.GetSymbol() if not number else atom.GetAtomicNum()
    return one_hot_encoding(atom_val, allowable_set, encode_unknown)


def atom_number(atom: dm.Atom):
    """Get the atomic number for an atom.

    Args:
        atom : RDKit atom instance.

    Returns:
        list: List containing one int only.

    """
    return [atom.GetAtomicNum()]


def atom_degree_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the degree of an atom.

    Result might be different if hydrogen have been added or not.

    Args:
        atom: RDKit atom instance.
        allowable_set : Atom degrees to consider. Default: ATOM_DEGREE_LIST
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.

    """
    if allowable_set is None:
        allowable_set = ATOM_DEGREE_LIST
    return one_hot_encoding(atom.GetDegree(), allowable_set, encode_unknown)


def atom_degree(
    atom: dm.Atom,
):
    """Get the degree of an atom.

    Result might be different if hydrogen have been added or not.

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one int only.
    """
    return [atom.GetDegree()]


def atom_total_degree_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the total degree of an atom including Hs.

    Args:
        atom: RDKit atom instance.
        allowable_set : Atom total degrees to consider. Default: ATOM_TOTAL_DEGREE_LIST
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.

    """
    if allowable_set is None:
        allowable_set = ATOM_TOTAL_DEGREE_LIST
    return one_hot_encoding(atom.GetTotalDegree(), allowable_set, encode_unknown)


def atom_total_degree(atom: dm.Atom):
    """Get the total degree of an atom.

    Result might be different if hydrogen have been added or not.

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one int only.
    """
    return [atom.GetTotalDegree()]


def atom_explicit_valence_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the explicit valence of an aotm.

    Args:
        atom: RDKit atom instance.
        allowable_set : Atom explicit valence to consider. Default: EXPLICIT_VALENCE_LIST
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = EXPLICIT_VALENCE_LIST
    return one_hot_encoding(atom.GetExplicitValence(), allowable_set, encode_unknown)


def atom_explicit_valence(atom: dm.Atom):
    """Get the explicit valence of an atom.

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one int only.
    """
    return [atom.GetExplicitValence()]


def atom_implicit_valence_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the implicit valence of an aotm.

    Args:
        atom: RDKit atom instance.
        allowable_set : Atom implicit valence to consider. Default: IMPLICIT_VALENCE_LIST
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = IMPLICIT_VALENCE_LIST
    return one_hot_encoding(atom.GetImplicitValence(), allowable_set, encode_unknown)


def atom_implicit_valence(atom: dm.Atom):
    """Get the implicit valence of an atom.

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one int only.
    """
    return [atom.GetImplicitValence()]


def atom_hybridization_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the hybridization of an atom.

    Args:
        atom: RDKit atom instance.
        allowable_set : Atom hybridization state to consider. Default: HYBRIDIZATION_LIST
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = HYBRIDIZATION_LIST
    return one_hot_encoding(atom.GetHybridization(), allowable_set, encode_unknown)


def atom_total_num_H_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the total number of Hs of an atom.

     Args:
        atom: RDKit atom instance.
        allowable_set : Atom total hydrogen to consider. Default: ATOM_NUM_H_LIST
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = ATOM_NUM_H_LIST
    return one_hot_encoding(atom.GetTotalNumHs(), allowable_set, encode_unknown)


def atom_total_num_H(atom: dm.Atom):
    """Get the total number of Hs of an atom.

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one int only.
    """
    return [atom.GetTotalNumHs()]


def atom_formal_charge_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the formal charge of an atom.

     Args:
        atom: RDKit atom instance.
        allowable_set : Atom formal charge to consider. Default: HYBRIDIZATION_LIST
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = CHARGE_LIST
    return one_hot_encoding(atom.GetFormalCharge(), allowable_set, encode_unknown)


def atom_formal_charge(atom: dm.Atom):
    """Get formal charge for an atom.

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one int only.
    """
    return [atom.GetFormalCharge()]


def atom_partial_charge(atom: dm.Atom):
    """Get Gasteiger partial charge for an atom.

    This function requires calling ``AllChem.ComputeGasteigerCharges(mol)`` first to compute Gasteiger charges.
    Nan and infinite values are set to 0

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one float only.
    """
    if not atom.HasProp("_GasteigerCharge"):
        mol = atom.GetOwningMol()
        AllChem.ComputeGasteigerCharges(mol)
    gasteiger_charge = atom.GetProp("_GasteigerCharge")
    gasteiger_charge = np.nan_to_num(float(gasteiger_charge), posinf=0.0, neginf=0.0)
    return [float(gasteiger_charge)]


def atom_num_radical_electrons_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the number of radical electrons of an atom.

     Args:
        atom: RDKit atom instance.
        allowable_set : Atom num radical electron to consider. Default: RADICAL_ELECTRON_LIST
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = RADICAL_ELECTRON_LIST
    return one_hot_encoding(atom.GetNumRadicalElectrons(), allowable_set, encode_unknown)


def atom_num_radical_electrons(atom: dm.Atom):
    """Get the number of radical electrons for an atom.

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one int only.
    """
    return [atom.GetNumRadicalElectrons()]


def atom_is_aromatic_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for whether the atom is aromatic.

     Args:
        atom: RDKit atom instance.
        allowable_set : Atom aromatic state to consider. Default: [True, False]
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = [False, True]
    return one_hot_encoding(atom.GetIsAromatic(), allowable_set, encode_unknown)


def atom_is_aromatic(atom: dm.Atom):
    """Get whether the atom is aromatic.


    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one bool only.
    """
    return [atom.GetIsAromatic()]


def atom_is_in_ring_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for whether the atom is in ring.

     Args:
        atom: RDKit atom instance.
        allowable_set : Atom ring state to consider. Default: [True, False]
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = [False, True]
    return one_hot_encoding(atom.IsInRing(), allowable_set, encode_unknown)


def atom_is_in_ring(atom: dm.Atom):
    """Get whether the atom is in ring.

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one int only.
    """
    return [atom.IsInRing()]


def atom_chiral_tag_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the chiral tag of an atom.

     Args:
        atom: RDKit atom instance.
        allowable_set : Atom chiral tag to consider. Default: CHIRAL_TYPES
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    list
        list: List of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = CHIRAL_TYPES
    return one_hot_encoding(atom.GetChiralTag(), allowable_set, encode_unknown)


def atom_chirality_type_one_hot(
    atom: dm.Atom,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the chirality type of an atom.

     Args:
        atom: RDKit atom instance.
        allowable_set : Atom chirality to consider. Default: ["R", "S"]
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    list
        list: List of boolean values where at most one value is True.
    """

    cip_code = None
    if atom.HasProp("_CIPCode"):
        cip_code = atom.GetProp("_CIPCode")
    if allowable_set is None:
        allowable_set = ["R", "S"]
    return one_hot_encoding(cip_code, allowable_set, encode_unknown)


def atom_mass(atom: dm.Atom, coeff: float = 1):
    """Get the mass of an atom and scale it.

    Args:
        atom: RDKit atom instance.
        coeff: scaling factor for the atom mass
    Returns:
        list: List containing one float only.
    """
    return [atom.GetMass() * coeff]


def atom_is_chiral_center(atom: dm.Atom):
    """Get whether the atom is chiral center

    Args:
        atom: RDKit atom instance.
    Returns:
        list: List containing one bool only.
    """
    return [atom.HasProp("_ChiralityPossible")]


def atom_extended_properties(atom: dm.Atom, dataset: str = "origin"):
    """Get a full set of atom descriptors

    Args:
        atom: RDKit atom instance.
        dataset: which dataset to use to query the atom descriptor.
            One of 'origin', 'elements', 'elements_completed'. Default to "origin".
    Returns:
        list: List containing one float only.
    """
    if dataset == "origin":
        data = get_data("origin")
    elif dataset in ["elements", "element"]:
        data = get_data("elements")
    else:
        data = get_data("elements_completed")

    feat_shape = data.shape[-1]
    feat = [0 for _ in range(feat_shape)]
    if atom.GetSymbol() in list(data.index):
        feat = list(data.loc[atom.GetSymbol()])
    return feat


def bond_type_one_hot(
    bond: dm.Bond,
    allowable_set: Optional[List[Any]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the type of a bond.

    Args:
        bond: RDKit atom instance.
        allowable_set : Bond type to consider. Default: BOND_TYPES
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: ist of boolean values where at most one value is True.

    """
    if allowable_set is None:
        allowable_set = BOND_TYPES
    return one_hot_encoding(bond.GetBondType(), allowable_set, encode_unknown)


def bond_is_conjugated_one_hot(
    bond: dm.Bond,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for whether the bond is conjugated.

    Args:
        bond: RDKit bond instance.
        allowable_set : Bond type to consider. Default: BOND_TYPES
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: list of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = [False, True]
    return one_hot_encoding(bond.GetIsConjugated(), allowable_set, encode_unknown)


def bond_is_conjugated(bond: dm.Bond):
    """Get whether the bond is conjugated.

    Args:
        bond: RDKit bond instance.
    Returns:
        list: List containing one bool only.
    """
    return [bond.GetIsConjugated()]


def bond_is_in_ring_one_hot(
    bond: dm.Bond,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for whether the bond is in a ring of any size.

    Args:
        bond: RDKit bond instance.
        allowable_set : Bond type to consider. Default: [True, False]
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: list of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = [False, True]
    return one_hot_encoding(bond.IsInRing(), allowable_set, encode_unknown)


def bond_is_in_ring(bond: dm.Bond):
    """Get whether the bond is in ring.

    Args:
        bond: RDKit bond instance.
    Returns:
        list: List containing one bool only.
    """
    return [bond.IsInRing()]


def bond_stereo_one_hot(
    bond: dm.Bond,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """One hot encoding for the stereo configuration of a bond.

    Args:
        bond: RDKit bond instance.
        allowable_set : Bond type to consider. Default: BOND_STEREO
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: list of boolean values where at most one value is True.
    """
    if allowable_set is None:
        allowable_set = BOND_STEREO
    return one_hot_encoding(bond.GetStereo(), allowable_set, encode_unknown)


def bond_direction_one_hot(
    bond: dm.Bond,
    allowable_set: Optional[List[str]] = None,
    encode_unknown: bool = False,
):
    """Get a one hot encoding for the direction of a bond.

    Args:
        bond: RDKit bond instance.
        allowable_set : Bond type to consider. Default: BOND_DIRECTION
        encode_unknown: whether to map inputs not in the allowable set to the additional last element. (Default: False)

    Returns:
        list: List of boolean values where at most one value is True.

    """
    if allowable_set is None:
        allowable_set = BOND_DIRECTION
    return one_hot_encoding(bond.GetBondDir(), allowable_set, encode_unknown)


def pairwise_2D_dist(mol: dm.Mol):
    """Compute the pairwise distance between all pairs of atoms in 2D

    Args:
        mol: Input molecule

    Returns:
        dist: Matrix of size V^2, 1
    """
    mat = GetDistanceMatrix(mol)
    return mat.reshape(-1, 1)


@requires_conformer
def pairwise_3D_dist(mol: dm.Mol, conformer_id: int = -1):
    """Compute the pairwise 3D distance between all pair of atoms

    Args:
        mol: Input molecule
        conformer_id: conformer id to use

    Returns:
        dist: Matrix of size V^2, 1
    """
    mat = Get3DDistanceMatrix(mol, confId=conformer_id)
    return mat.reshape(-1, 1)


def pairwise_dist_indicator(mol: dm.Mol, max_distance: int = 7):
    """Compute the pairwise distance matrix gated by a max distance for the weave featurizer

    Args:
        mol: input molecule
        max_distance: maximum distance to use for the thresholding
    """
    dist_mat = torch.from_numpy(pairwise_2D_dist(mol))
    # Elementwise compare if distance is bigger than 0, 1, ..., max_distance - 1
    dist_indicator = (dist_mat > torch.arange(0, max_distance).float()).float()
    return dist_indicator.numpy()


def pairwise_bond_indicator(mol: dm.Mol, allowable_set: Optional[List[Any]] = None):
    """Compute the pairwise bond indicator for weave net

    Args:
        mol: input molecule
        allowable_set: bond type to consider
    """
    num_atoms = mol.GetNumAtoms()
    if allowable_set is None:
        allowable_set = BOND_TYPES
    bond_indicators = torch.zeros(num_atoms, num_atoms, len(allowable_set))
    for bond in mol.GetBonds():
        bond_type_encoding = torch.tensor(
            bond_type_one_hot(bond, allowable_set=allowable_set)
        ).float()
        begin_atom_idx, end_atom_idx = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
        bond_indicators[begin_atom_idx, end_atom_idx] = bond_type_encoding
        bond_indicators[end_atom_idx, begin_atom_idx] = bond_type_encoding
    # Reshape from (V, V, num_bond_types) to (V^2, num_bond_types)
    bond_indicators = bond_indicators.reshape(-1, len(allowable_set))
    return bond_indicators.numpy()


def pairwise_ring_membership(mol: dm.Mol):
    """Compute the joint ring membership of all atom pairs

    Args:
        mol: Input molecule

    Returns:
        ring_membership: Matrix of size V^2, 1
    """
    num_atoms = mol.GetNumAtoms()
    sssr = GetSymmSSSR(mol)
    ring_mate = np.zeros((num_atoms, num_atoms, 1))
    for ring in sssr:
        ring = np.asarray(ring)
        num_atoms_in_ring = len(ring)
        for i in range(num_atoms_in_ring):
            ring_mate[ring[i], ring] = 1
    ring_mate = ring_mate.reshape(-1, 1)
    return ring_mate



from typing import List
from typing import Optional
from typing import Callable

import torch

from torch import nn
from loguru import logger


def get_default_hgf_pooler(name, config, **kwargs):
    """Get default HuggingFace pooler based on the model name
    Args:
        name: name of the model
        config: config of the model
        kwargs: additional arguments to pass to the pooler
    """
    model_type = config.get("model_type", None)
    if name not in ["bert", "roberta", "gpt", "bart"] and name in Pooling.SUPPORTED_POOLING[:-1]:
        return HFPooler(config, name=name, **kwargs)
    names = [name]
    if model_type is not None:
        names += [model_type]
    if any(x in ["bert", "roberta"] for x in names):
        return BertPooler(config, **kwargs)
    elif any(x.startswith("gpt") for x in names):
        return GPTPooler(config, **kwargs)
    elif any(x == "bart" for x in names):
        return BartPooler(config, **kwargs)
    return None


class Pooling(nn.Module):
    """
    Perform simple pooling on a tensor over one dimension
    """

    SUPPORTED_POOLING = ["mean", "avg", "max", "sum", "clf", None]

    def __init__(self, dim: int = 1, name: str = "max"):
        """
        Pooling for embeddings

        Args:
            dim: dimension to pool over, default is 1
            name: pooling type. Default is 'mean'.
        """
        super().__init__()
        self.dim = dim
        self.name = name

    def forward(self, x, indices: List[int] = None, mask: torch.Tensor = None) -> torch.Tensor:
        """Perform a pooling operation on the input tensor

        Args:
            x: input tensor to pull over
            indices: Subset of indices to pool over. Defaults to None for all indices.
            mask: binary mask to apply when pooling. Defaults to None, which is a matrix of 1.
                If mask is provided it takes precedence over indices.
        """
        x = torch.as_tensor(x)
        if mask is None:
            mask = torch.ones_like(x)
        if indices is not None:
            mask[:, indices] = 0
        neg_inf = torch.finfo(x.dtype).min
        if mask.ndim == 2:
            mask = mask.unsqueeze(-1)  # B, S, 1
        if self.name == "clf":
            return x[:, 0, :]
        if self.name == "max":
            tmp = x.masked_fill(mask, neg_inf)
            return torch.max(tmp, dim=self.dim)[0]
        elif self.name in ["mean", "avg"]:
            return torch.sum(x * mask, dim=self.dim) / mask.sum(self.dim)
        elif self.name == "sum":
            return torch.sum(x * mask, dim=self.dim)
        return x


class HFPooler(nn.Module):
    """Default Pooler based on Molfeat Pooling layer"""

    def __init__(self, config, dim: int = 1, name: str = "mean", **kwargs):
        super().__init__()
        self.config = config
        self.pooling = Pooling(dim=dim, name=name)

    def forward(
        self,
        h: torch.Tensor,
        inputs: Optional[torch.Tensor] = None,
        mask: Optional[torch.Tensor] = None,
        ignore_padding: bool = True,
        **kwargs,
    ) -> torch.Tensor:
        """Forward pass of the pooling layer

        Args:
            h: hidden representation of the input sequence to pool over
            inputs: optional input that has been provided to the underlying bert model
            mask: optional mask to use in place of computing the padding specific mask
            ignore_padding: whether to ignore padding tokens when pooling

        Returns:
            pooled_output: pooled representation of the input sequence
        """
        if mask is None and ignore_padding:
            mask = inputs.ne(self.config.get("pad_token_id"))
        if mask.ndim == 2:
            mask = mask.unsqueeze(-1)  # B, S, 1
        return self.pooling(h, indices=None, mask=mask)


class BertPooler(nn.Module):
    """
    Default Bert pooler as implemented in huggingface transformers
    The bert pooling function focuses on a projection of the first token ([CLS]) to get a sentence representation.
    """

    def __init__(
        self,
        config,
        activation: Optional[Callable] = None,
        random_seed: int = None,
        **kwargs,
    ):
        super().__init__()
        self.config = config
        self.random_seed = random_seed
        if self.random_seed is not None:
            torch.manual_seed(self.random_seed)
        hidden_size = config.get("hidden_size")
        self.dense = nn.Linear(hidden_size, hidden_size)
        self.activation = nn.Tanh() if activation is None else activation

    def forward(
        self, h: torch.Tensor, inputs: Optional[torch.Tensor] = None, **kwargs
    ) -> torch.Tensor:
        """Forward pass of the pooling layer

        Args:
            h: hidden representation of the input sequence to pool over
            inputs: optional input that has been provided to the underlying bert model

        Returns:
            pooled_output: pooled representation of the input sequence
        """
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token.
        first_token_tensor = h[:, 0]
        pooled_output = self.dense(first_token_tensor)
        pooled_output = self.activation(pooled_output)
        return pooled_output


class BartPooler(nn.Module):
    """
    Default Bart pooler as implemented in huggingface transformers
    The Bart pooling function focusing on the eos token ([EOS]) to get a sentence representation.
    """

    def __init__(self, config, **kwargs):
        super().__init__()
        self.config = config

    def forward(
        self, h: torch.Tensor, inputs: Optional[torch.Tensor] = None, **kwargs
    ) -> torch.Tensor:
        """Forward pass of the pooling layer

        Args:
            h: hidden representation of the input sequence to pool over
            inputs: inputs tokens to the bart underlying model

        Returns:
            pooled_output: pooled representation of the input sequence
        """
        eos_mask = inputs.eq(self.config.get("eos_token_id"))
        if len(torch.unique_consecutive(eos_mask.sum(1))) > 1:
            raise ValueError("All examples must have the same number of <eos> tokens.")
        pooled_output = h[eos_mask, :].view(h.size(0), -1, h.size(-1))[:, -1, :]
        return pooled_output


class GPTPooler(nn.Module):
    """
    Default GPT pooler as implemented in huggingface transformers
    The GPT pooling function focusing on the last non-padding token given sequence length to get a sentence representation.
    """

    def __init__(self, config, **kwargs):
        super().__init__()
        self.config = config
        self.pad_token_id = config.get("pad_token_id")

    def forward(
        self, h: torch.Tensor, inputs: Optional[torch.Tensor] = None, **kwargs
    ) -> torch.Tensor:
        """Forward pass of the pooling layer

        Args:
            h: hidden representation of the input sequence to pool over
            inputs: inputs tokens to the bart underlying model

        Returns:
            pooled_output: pooled representation of the input sequence
        """
        batch_size, sequence_lengths = inputs.shape[:2]

        assert (
            self.pad_token_id is not None or batch_size == 1
        ), "Cannot handle batch sizes > 1 if no padding token is defined."
        if self.pad_token_id is None:
            sequence_lengths = -1
            logger.warning(
                f"{self.__class__.__name__} will not detect padding tokens in `inputs_embeds`. Results may be "
                f"unexpected if using padding tokens in conjunction with `inputs_embeds.`"
            )
        else:
            sequence_lengths = torch.ne(inputs, self.pad_token_id).sum(-1) - 1
        pooled_output = h[torch.arange(batch_size), sequence_lengths]
        return pooled_output



import pytest

import numpy as np
import pandas as pd

from sklearn.pipeline import Pipeline
from sklearn.naive_bayes import GaussianNB
from sklearn.compose import ColumnTransformer
from molfeat.trans.base import MoleculeTransformer


@pytest.fixture
def smiles():
    return [
        "CC1CC2C3CCC4=CC(=O)C=CC4(C3(C(CC2(C(=O)CO1)O)C)O)C",
        "CN(CCOC(c1ccccc1)c1ccccc1)C",
        "O/N=C(/c1csc(n1)N)\C(=O)N[C@@H]1C(=O)N2[C@@H]1SCC(=C2C(=O)O)C=C",
        "CC(C)(C)NCC(C1=CC(=C(C=C1)O)CO)O",
    ]


@pytest.fixture(params=["list", "series", "dataframe"])
def mols(request, smiles):
    if request.param == "list":
        return smiles
    elif request.param == "series":
        return pd.Series(smiles, name="smiles")
    elif request.param == "dataframe":
        return pd.DataFrame({"smiles": smiles, "column_2": [1, 0, 1, 1]})


def test_list_series_dataframe(mols):
    transformer_ecfp = MoleculeTransformer(featurizer="ecfp", dtype=float)
    results = transformer_ecfp.fit_transform(mols)

    assert results.shape == (4, 2048)
    assert isinstance(results, np.ndarray)


def test_with_pipeline_column_transformer(smiles):
    # setup data
    mols = pd.DataFrame({"smiles": smiles, "column_2": [1, 0, 1, 1]})

    # setup pipeline
    transformer_ecfp = MoleculeTransformer(featurizer="ecfp", dtype=float)
    column_preprocessor = ColumnTransformer(
        transformers=[
            ("ecfp_trans", transformer_ecfp, ["smiles"]),
            ("column_2", "passthrough", ["column_2"]),
        ]
    )

    pipeline = Pipeline([("preprocess", column_preprocessor), ("classifier", GaussianNB())])

    # fit/predict pipeline
    pipeline.fit(mols, [1, 0, 1, 0])
    r = pipeline.predict(mols)

    # tests
    expect = np.ndarray
    assert isinstance(r, expect)

    expect = (4,)
    assert r.shape == expect



from molfeat.calc import Pharmacophore2D
from molfeat import viz


def test_colors_from_feature_factory():
    # Get a factory
    calc = Pharmacophore2D(factory="pmapper")
    feature_factory = calc.sig_factory.featFactory  # type: ignore

    # Get colors
    colors = viz.colors_from_feature_factory(feature_factory)

    # Check
    assert isinstance(colors, dict)



from inspect import isclass
from inspect import ismodule

from typing import TYPE_CHECKING
from typing import overload
from typing import Any
from typing import Union
from typing import Optional
from typing import Callable
from typing import Literal
from typing import NoReturn
from typing import Tuple
from typing import Type

import sys

if sys.version_info < (3, 10):
    from importlib_metadata import EntryPoint
else:
    from importlib.metadata import EntryPoint

from molfeat.plugins.exception import InvalidEntryPointTypeError

__all__ = (
    "CalculatorFactory",
    "TransformerFactory",
    "PretrainedTransformerFactory",
    "DefaultFactory",
)

if TYPE_CHECKING:
    from molfeat.calc import SerializableCalculator
    from molfeat.trans import MoleculeTransformer
    from molfeat.trans.pretrained import PretrainedMolTransformer


def raise_invalid_type_error(
    entry_point_name: str, entry_point_group: str, valid_classes: Tuple[Any, ...]
) -> NoReturn:
    """Raise an `InvalidEntryPointTypeError` with formatted message.

    Args:
        entry_point_name: name of the entry point
        entry_point_group: name of the entry point group
        valid_classes: tuple of valid classes for the given entry point group

    Raises:
        InvalidEntryPointTypeError: always
    """
    template = "entry point `{}` registered in group `{}` is invalid because its type is not one of the supported types ({})"
    args = (
        entry_point_name,
        entry_point_group,
        ", ".join([e.__name__ for e in valid_classes]),
    )
    raise InvalidEntryPointTypeError(template.format(*args))


def BaseFactory(group: str, name: str, load: bool = True) -> Union[EntryPoint, Any]:
    """Return the plugin class registered under a given entry point group and name.

    Args:
        group: entry point group
        name: entry point name
        load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
    Return:
        the plugin class
    Raises:
        MissingEntryPointError: entry point was not registered
        MultipleEntryPointError: entry point could not be uniquely resolved
        LoadingEntryPointError: entry point could not be loaded
    """
    # circular import
    from .entry_point import get_entry_point, load_entry_point

    if load is True:
        return load_entry_point(group, name)

    return get_entry_point(group, name)


@overload
def CalculatorFactory(
    entry_point_name: str,
    load: Literal[True] = True,
    entry_point_group: Optional[str] = None,
) -> Union[Type["SerializableCalculator"], Callable]:
    ...


@overload
def CalculatorFactory(
    entry_point_name: str, load: Literal[False], entry_point_group: Optional[str] = None
) -> EntryPoint:
    ...


def CalculatorFactory(
    entry_point_name: str,
    load: bool = True,
    entry_point_group: Optional[str] = None,
) -> Union[EntryPoint, Type["SerializableCalculator"], Callable]:
    """Return the `SerializableCalculator` sub class registered under the given entry point.

    Args:
        entry_point_name: the entry point name.
        load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
        entry_point_group: the optional entry point group to use

    Return:
        sub class of :py:class:`~molfeat.calc.SerializableCalculator`
    """
    from molfeat.calc import SerializableCalculator

    if entry_point_group is None:
        entry_point_group = "molfeat.calc"
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (SerializableCalculator,)

    if not load:
        return entry_point

    # if the entry point is a module, nothing to do
    if ismodule(entry_point):
        return entry_point
    if isclass(entry_point) and issubclass(entry_point, valid_classes):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def TransformerFactory(
    entry_point_name: str,
    load: Literal[True] = True,
    entry_point_group: Optional[str] = None,
) -> Union[Type["MoleculeTransformer"], Callable]:
    ...


@overload
def TransformerFactory(
    entry_point_name: str, load: Literal[False], entry_point_group: Optional[str] = None
) -> EntryPoint:
    ...


def TransformerFactory(
    entry_point_name: str,
    load: bool = True,
    entry_point_group: Optional[str] = None,
) -> Union[EntryPoint, Type["MoleculeTransformer"], Callable]:
    """Return the `MoleculeTransformer` sub class registered under the given entry point.

    Args:
        entry_point_name: the entry point name.
        load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
        entry_point_group: the optional entry point group to use

    Returns:
        sub class of :py:class:`~molfeat.trans.MoleculeTransformer`

    Raise:
        InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from molfeat.trans import MoleculeTransformer
    from molfeat.trans import BaseFeaturizer

    if entry_point_group is None:
        entry_point_group = "molfeat.trans"
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (MoleculeTransformer, BaseFeaturizer)

    if not load:
        return entry_point

    # if the entry point is a module, nothing to do
    if ismodule(entry_point):
        return entry_point
    if isclass(entry_point) and issubclass(entry_point, valid_classes):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def PretrainedTransformerFactory(
    entry_point_name: str,
    load: Literal[True] = True,
    entry_point_group: Optional[str] = None,
) -> Union[Type["PretrainedMolTransformer"], Callable]:
    ...


@overload
def PretrainedTransformerFactory(entry_point_name: str, load: Literal[False]) -> EntryPoint:
    ...


def PretrainedTransformerFactory(
    entry_point_name: str,
    load: bool = True,
    entry_point_group: Optional[str] = None,
) -> Union[EntryPoint, Type["PretrainedMolTransformer"], Callable]:
    """Return the PretrainedMolTransformer sub class registered under the given entry point.

    Args:
        entry_point_name: the entry point name.
        load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
        entry_point_group: the optional entry point group to use

    Returns:
        sub class of :py:class:`~molfeat.trans.pretrained.PretrainedMolTransformer`

    Raise:
        InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """
    from molfeat.trans import MoleculeTransformer
    from molfeat.trans.pretrained import PretrainedMolTransformer

    if entry_point_group is None:
        entry_point_group = "molfeat.trans.pretrained"
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)
    valid_classes = (PretrainedMolTransformer, MoleculeTransformer)

    if not load:
        return entry_point
    # if the entry point is a module, nothing to do
    if ismodule(entry_point):
        return entry_point
    if isclass(entry_point) and issubclass(entry_point, valid_classes):
        return entry_point

    raise_invalid_type_error(entry_point_name, entry_point_group, valid_classes)


@overload
def DefaultFactory(
    entry_point_name: str,
    load: Literal[True] = True,
    entry_point_group: str = None,
) -> Union[Type["PretrainedMolTransformer"], Callable]:
    ...


@overload
def DefaultFactory(
    entry_point_name: str, load: Literal[False], entry_point_group: str = None
) -> EntryPoint:
    ...


def DefaultFactory(
    entry_point_name: str,
    load: bool = True,
    entry_point_group: str = None,
) -> Union[EntryPoint, Type["PretrainedMolTransformer"], Callable]:
    """Return the Default factory for extending capabilities given a specific module.

    Args:
        entry_point_name: the entry point name.
        load: if True, load the matched entry point and return the loaded resource instead of the entry point itself.
        entry_point_group: the optional entry point group to use

    Returns:
        sub class or module of

    Raise:
        InvalidEntryPointTypeError: if the type of the loaded entry point is invalid.
    """

    if entry_point_group is None:
        entry_point_group = "molfeat"
    entry_point = BaseFactory(entry_point_group, entry_point_name, load=load)

    if not load:
        return entry_point
    # if the entry point is a module, nothing to do
    if ismodule(entry_point):
        return entry_point
    raise_invalid_type_error(entry_point_name, entry_point_group, ())



try:
    from importlib.metadata import version
    from importlib.metadata import PackageNotFoundError
except ModuleNotFoundError:
    # Try backported to PY<38 `importlib_metadata`.
    from importlib_metadata import version
    from importlib_metadata import PackageNotFoundError

try:
    __version__ = version("molfeat")
except PackageNotFoundError:
    # package is not installed
    __version__ = "dev"



from typing import Union
from typing import Optional

import itertools
import numpy as np
import pandas as pd
import datamol as dm

from collections import defaultdict as ddict
from pathlib import Path
from loguru import logger
from rdkit import Chem
from rdkit.Chem import rdMolDescriptors as Desc
from rdkit.Chem.rdmolops import GetFormalCharge
from rdkit.Chem.Scaffolds import MurckoScaffold
from molfeat.calc.base import SerializableCalculator

import molfeat


def _is_ring_fully_conjugated(mol: dm.Mol, ring: list):
    """Check whether a ring is fully conjugated"""
    suppl = Chem.ResonanceMolSupplier(mol)
    first_idx_conj = -1
    first = True
    for i in range(len(ring)):
        atom_idx = ring[i]
        atom_conj_grp_idx = Chem.ResonanceMolSupplier.GetAtomConjGrpIdx(suppl, atom_idx)
        if first:
            first_idx_conj = atom_conj_grp_idx
            if first_idx_conj > 99999 or first_idx_conj == -1:
                return False
            first = False
        else:
            if atom_conj_grp_idx != first_idx_conj:
                return False
    return True


def _n_multiple_bond_in_ring(mol: dm.Mol, ring: list):
    """Get number of multiple bonds in a ring"""
    atom_i = -1
    atom_j = -1
    bond = None
    nr_multiple_bonds = 0
    multiple_bond_types = [2, 3, 12]
    for i in range(len(ring)):
        atom_i = ring[i]
        for j in range(0, i):
            atom_j = ring[j]
            bond = mol.GetBondBetweenAtoms(atom_i, atom_j)
            if bond is not None:
                bond_type = Chem.Bond.GetBondType(bond)
                if int(bond_type) in multiple_bond_types:
                    nr_multiple_bonds += 1
    return nr_multiple_bonds


def _count_heteroatom_per_ring(mol: dm.Mol, rings: list):
    """Count number of heteroatoms in rings"""
    n_heteros = [0] * len(rings)
    for i, ring in enumerate(rings):
        for atom in ring:
            is_hetero = int(mol.GetAtomWithIdx(atom).GetAtomicNum() not in [1, 6])
            n_heteros[i] += is_hetero
    return n_heteros


def _get_ring_system(mol: dm.Mol):
    """Build ring systems in a molecule"""
    q = mol.GetRingInfo()
    simple_rings = q.AtomRings()
    rings = [set(r) for r in simple_rings]
    ring_map = [set([x]) for x in range(len(rings))]
    go_next = True
    while go_next:
        go_next = False
        for i, j in itertools.combinations(range(len(rings)), 2):
            if rings[i] & rings[j]:
                new_map = set().union(ring_map[i], ring_map[j])
                q = rings[i] | rings[j]
                min_ind, max_ind = min(i, j), max(i, j)
                del rings[max_ind], rings[min_ind]
                del ring_map[max_ind], ring_map[min_ind]
                rings.append(q)
                ring_map.append(new_map)

                go_next = True
                break
    return list(simple_rings), rings, ring_map


def _ring_atom_state(mol: dm.Mol):
    """Get the conjugated state of ring atoms"""
    ri = mol.GetRingInfo()
    ring_atom_conj_state = ddict(list)
    for ring in ri.AtomRings():
        state = _is_ring_fully_conjugated(mol, ring)
        for atom in ring:
            ring_atom_conj_state[atom].append(state)
    return ring_atom_conj_state


class ScaffoldKeyCalculator(SerializableCalculator):
    """
    Implementation of the Scaffold Keys described in
    `Identification of Bioisosteric Scaffolds using Scaffold Keys` by Peter Ertl
    """

    DESCRIPTORS = [
        "n_atom_in_rings",
        "n_atom_in_conjugated_ring",
        "n_atoms_not_in_conjugated_ring",
        "n_atom_in_chain",
        "n_atom_exocyclic",
        "n_nitrogen",
        "n_nitrogen_in_ring",
        "n_oxygen",
        "n_oxygen_in_ring",
        "n_sulfur",
        "n_heteroatoms",
        "n_heteroatoms_in_ring",
        "n_atom_spiro_atoms",
        "n_heteroatom_more_than_2_conn",
        "n_carbon_atleast_2_heteroatoms",
        "n_atom_at_least_2_nei_more_than_2_conn",
        "abs_scaffold_format_charge",
        "n_bonds",
        "n_multiple_non_conj_ring_bonds",
        "n_bonds_2_heteroatoms",
        "n_carbon_het_carbon_het_bonds",
        "n_bonds_at_least_3_conn",
        "n_exocyclic_single_bonds_carbon",
        "n_exocyclic_single_bonds_nitrogen",
        "n_non_ring_bonds_2_conj_rings",
        "n_non_ring_bonds_conj_nonconj_rings",
        "n_bonds_atoms_with_at_least_one_nei_with_2_conn",
        "n_simple_rings",
        "size_largest_ring",
        "n_simple_rings_no_heteroatoms",
        "n_simple_rings_1_heteroatoms",
        "n_simple_rings_2_heteroatoms",
        "n_simple_rings_at_least_3_heteroatoms",
        "n_simple_non_conj_5_atoms_rings",
        "n_simple_non_conj_6_atoms_rings",
        "n_ring_system",
        "n_ring_system_with_2_non_conj_simple_ring",
        "n_ring_system_with_2_conj_simple_ring",
        "n_ring_system_with_conj_non_conj_simple_ring",
        "n_ring_system_with_3_conj_simple_ring",
        "n_ring_system_with_3_non_conj_simple_ring",
        "n_ring_system_with_greater_one_conj_nonconj_simple_ring",
    ]

    NORM_PARAMS = pd.read_csv(
        Path(molfeat.__file__).parents[0].joinpath("data/skey_parameters.csv"),
        index_col=0,
    ).loc[DESCRIPTORS]

    def __init__(
        self, normalize: bool = False, verbose: bool = False, use_scaffold: bool = False, **kwargs
    ):
        """
        Init of the scaffold key function

        Args:
            normalize: whether to normalize the value of the feature
            verbose: whether to log errors
            use_scaffold: whether to convert the molecule into scaffold first
        """
        self.normalize = normalize
        self.verbose = verbose
        self.use_scaffold = use_scaffold

    def __getstate__(self):
        """Get state of the scaffold key function"""
        state = {}
        state["normalize"] = self.normalize
        state["verbose"] = self.verbose
        state["use_scaffold"] = self.use_scaffold
        return state

    def __len__(self):
        return len(self.DESCRIPTORS)

    @classmethod
    def compute_normalization(cls, features: np.ndarray):
        """Normalize input features. The normalization parameters are
        computed by the scaffolds of 2.1M molecules from CHEMBL 29.
        """
        return (features - cls.NORM_PARAMS["mean"]) / cls.NORM_PARAMS["std"]

    def n_atom_in_rings(self, mol: dm.Mol):
        """1. number of ring atoms"""
        sm = dm.from_smarts("[r]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_atom_in_conjugated_ring(self, mol: dm.Mol):
        """2. number of atoms in conjugated rings"""
        ri = mol.GetRingInfo()
        n = 0
        for ring in ri.AtomRings():
            if _is_ring_fully_conjugated(mol, ring):
                n += len(ring)
        return n

    def n_atoms_not_in_conjugated_ring(self, mol: dm.Mol):
        """
        3. number of atoms not in conjugated rings
        (i.e. atoms in aliphatic rings and non-ring atoms)
        """
        # EN: replace conjugation by aromatic
        ri = mol.GetRingInfo()
        n = 0
        for ring in ri.AtomRings():
            if not _is_ring_fully_conjugated(mol, ring):
                n += len(ring)
        return n

    def n_atom_in_chain(self, mol: dm.Mol):
        """4. number atoms in chains (not counting double-connected exo-chain atoms)"""
        sm = dm.from_smarts("[!r;!$(*=[r])]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_atom_exocyclic(self, mol: dm.Mol):
        """5. number of exocyclic atoms (connected by multiple bonds to a ring)"""
        sm = dm.from_smarts("[!r;!$(*-[r])&$(*~[r])]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_nitrogen(self, mol: dm.Mol):
        """6. number of nitrogen"""
        sm = dm.from_smarts("[#7]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_nitrogen_in_ring(self, mol: dm.Mol):
        """7. number of nitrogen in rings"""
        sm = dm.from_smarts("[#7;r]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_oxygen(self, mol: dm.Mol):
        """8. number of oxygen"""
        sm = dm.from_smarts("[#8]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_oxygen_in_ring(self, mol: dm.Mol):
        """9. number of oxygen in rings"""
        sm = dm.from_smarts("[#8]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_sulfur(self, mol: dm.Mol):
        """10. number of sulfur atoms"""
        sm = dm.from_smarts("[#16]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_heteroatoms(self, mol: dm.Mol):
        """11. number of heteroatoms"""

        sm = dm.from_smarts("[!#1&!#6]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_heteroatoms_in_ring(self, mol: dm.Mol):
        """12. number of heteroatoms in rings"""
        sm = dm.from_smarts("[!#1&!#6&r]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_atom_spiro_atoms(self, mol: dm.Mol):
        """13. number of spiro atoms"""
        return Desc.CalcNumSpiroAtoms(mol)

    def n_heteroatom_more_than_2_conn(self, mol: dm.Mol):
        """14. number of heteroatoms with more than 2 connections"""
        sm = dm.from_smarts("[!#1;!#6;!D1!D0;!D2]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_carbon_atleast_2_heteroatoms(self, mol: dm.Mol):
        """15. number of carbon atoms connected to at least 2 heteroatoms"""
        n_atoms = 0
        for atom in mol.GetAtoms():
            tmp = [x for x in atom.GetNeighbors() if x.GetAtomicNum() not in [1, 6]]
            n_atoms += len(tmp) >= 2
        return n_atoms

    def n_atom_at_least_2_nei_more_than_2_conn(self, mol: dm.Mol):
        """16. Number of atoms where at least 2 connected atoms have more than 2 connections"""
        n_atoms = 0
        for atom in mol.GetAtoms():
            tmp = [x for x in atom.GetNeighbors() if len(x.GetNeighbors()) > 2]
            n_atoms += len(tmp) > 2
        return n_atoms

    def abs_scaffold_format_charge(self, mol: dm.Mol):
        """17. absolute value of the scaffold formal charge"""
        charge = GetFormalCharge(mol)
        return abs(charge)

    def n_bonds(self, mol: dm.Mol):
        """18. number of bonds"""
        return mol.GetNumBonds()

    def n_multiple_non_conj_ring_bonds(self, mol: dm.Mol):
        """19. number of multiple, nonconjugated ring bonds"""
        extracted_rings = []
        nr_multiple_bonds_infcr = 0  # infcr: in not fully conjugated ring
        rings = Chem.GetSymmSSSR(mol)
        for i in range(len(rings)):
            extracted_rings.append(list(rings[i]))
        for ring in extracted_rings:
            if not _is_ring_fully_conjugated(mol, ring):
                nr_multiple_bonds_infcr += _n_multiple_bond_in_ring(mol, ring)
        return nr_multiple_bonds_infcr

    def n_bonds_2_heteroatoms(self, mol: dm.Mol):
        """20. number of bonds connecting 2 heteroatoms"""
        sm = dm.from_smarts("[!#1&!#6]~[!#1&!#6]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_carbon_het_carbon_het_bonds(self, mol: dm.Mol):
        """21. number of bonds connecting 2 heteroatoms through 2 carbons"""
        sm = dm.from_smarts("[!#1&!#6]~[#6]~[#6]~[!#1&!#6]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_bonds_at_least_3_conn(self, mol: dm.Mol):
        """22. number of bonds with at least 3 connections on both its atoms"""
        sm = dm.from_smarts("[$([!#1](~[!#1])(~[!#1])~[!#1])][$([!#1](~[!#1])(~[!#1])~[!#1])]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_exocyclic_single_bonds_carbon(self, mol: dm.Mol):
        """23. number of exocyclic single bonds where a ring atom is carbon"""
        sm = dm.from_smarts("[!R;!#1]-[#6;R]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_exocyclic_single_bonds_nitrogen(self, mol: dm.Mol):
        """24. number of exocyclic single bonds where a ring atom is nitrogen"""
        sm = dm.from_smarts("[!R;!#1]-[#7;R]")
        return len(mol.GetSubstructMatches(sm, uniquify=True))

    def n_non_ring_bonds_2_conj_rings(self, mol: dm.Mol):
        """25. number of non-ring bonds connecting 2 nonconjugated rings"""
        # EN: this is interpretated literally as bonds and not path
        ring_atom_conj_state = _ring_atom_state(mol)
        sm = dm.from_smarts("[R:1]!@[R:2]")
        bond_list = mol.GetSubstructMatches(sm, uniquify=True)
        result = 0
        for a_start, a_end in bond_list:
            s_state = ring_atom_conj_state.get(a_start)
            e_state = ring_atom_conj_state.get(a_end)
            if False in s_state and False in e_state:
                result += 1
        return result

    def n_non_ring_bonds_conj_nonconj_rings(self, mol: dm.Mol):
        """
        26. number of non-ring bonds connecting 2 rings,
        one of them conjugated and one non-conjugated
        """
        # EN: this is interpretated literally as bonds and not path

        ring_atom_conj_state = _ring_atom_state(mol)
        sm = dm.from_smarts("[R:1]!@[R:2]")
        bond_list = mol.GetSubstructMatches(sm, uniquify=True)
        result = 0
        for a_start, a_end in bond_list:
            s_state = ring_atom_conj_state.get(a_start)
            e_state = ring_atom_conj_state.get(a_end)
            if (True in s_state and False in e_state) or (False in s_state and True in e_state):
                result += 1
        return result

    def n_bonds_atoms_with_at_least_one_nei_with_2_conn(self, mol: dm.Mol):
        """
        27. number of bonds where both atoms have at least one neighbor
        (not considering the bond atoms) with more than 2 connections
        """
        result = 0
        huge_conn = list(
            itertools.chain(*mol.GetSubstructMatches(dm.from_smarts("[*;!D0;!D1;!D2]"), uniquify=1))
        )
        for bond in mol.GetBonds():
            a_start, a_end = bond.GetBeginAtom(), bond.GetEndAtom()
            # we need to exclud the bond atom themselves
            allowed_conn_table = [
                x for x in huge_conn if x not in [a_start.GetIdx(), a_end.GetIdx()]
            ]
            if any([x.GetIdx() in allowed_conn_table for x in a_start.GetNeighbors()]) and any(
                [y.GetIdx() in allowed_conn_table for y in a_end.GetNeighbors()]
            ):
                result += 1
        return result

    def n_simple_rings(self, mol: dm.Mol):
        """28. number of simple rings"""
        ri = mol.GetRingInfo()
        return ri.NumRings()

    def size_largest_ring(self, mol: dm.Mol):
        """29. Size of the largest ring"""
        ri = mol.GetRingInfo()
        max_ring_size = max((len(r) for r in ri.AtomRings()), default=0)
        return max_ring_size

    def n_simple_rings_no_heteroatoms(self, mol: dm.Mol):
        """30. number of simple rings with no heteroatoms"""
        ri = mol.GetRingInfo()
        n_heteros = _count_heteroatom_per_ring(mol, ri.AtomRings())
        return sum(1 for x in n_heteros if x == 0)

    def n_simple_rings_1_heteroatoms(self, mol: dm.Mol):
        """31. number of simple rings with 1 heteroatom"""
        ri = mol.GetRingInfo()
        n_heteros = _count_heteroatom_per_ring(mol, ri.AtomRings())
        return sum(1 for x in n_heteros if x == 1)

    def n_simple_rings_2_heteroatoms(self, mol: dm.Mol):
        """32. number of simple rings with 2 heteroatom"""
        ri = mol.GetRingInfo()
        n_heteros = _count_heteroatom_per_ring(mol, ri.AtomRings())
        return sum(1 for x in n_heteros if x == 2)

    def n_simple_rings_at_least_3_heteroatoms(self, mol: dm.Mol):
        """33. number of simple rings with 3 or more heteroatoms"""
        ri = mol.GetRingInfo()
        n_heteros = _count_heteroatom_per_ring(mol, ri.AtomRings())
        return sum(1 for x in n_heteros if x >= 3)

    def n_simple_non_conj_5_atoms_rings(self, mol: dm.Mol):
        """34. number of simple non-conjugated rings with 5 atoms"""
        ri = mol.GetRingInfo()
        n = 0
        for ring in ri.AtomRings():
            if not _is_ring_fully_conjugated(mol, ring) and len(ring) == 5:
                n += 1
        return n

    def n_simple_non_conj_6_atoms_rings(self, mol: dm.Mol):
        """35. number of simple non-conjugated rings with 6 atoms"""
        ri = mol.GetRingInfo()
        n = 0
        for ring in ri.AtomRings():
            if not _is_ring_fully_conjugated(mol, ring) and len(ring) == 6:
                n += 1
        return n

    def n_ring_system(self, mol: dm.Mol):
        """36. number of ring systems"""
        simple_rings, ring_system, _ = _get_ring_system(mol)
        return len(ring_system)

    def n_ring_system_with_2_non_conj_simple_ring(self, mol: dm.Mol):
        """37. number of rings systems with 2 non-conjugated simple rings"""
        simple_rings, _, ring_map = _get_ring_system(mol)
        conj_rings_map = dict(
            (i, _is_ring_fully_conjugated(mol, x)) for i, x in enumerate(simple_rings)
        )
        result = 0
        for ring_set in ring_map:
            n_not_conj = sum(not conj_rings_map[rnum] for rnum in ring_set)
            result += n_not_conj == 2
        return result

    def n_ring_system_with_2_conj_simple_ring(self, mol: dm.Mol):
        """38. number of rings systems with 2 conjugated simple rings"""
        simple_rings, _, ring_map = _get_ring_system(mol)
        conj_rings_map = dict(
            (i, _is_ring_fully_conjugated(mol, x)) for i, x in enumerate(simple_rings)
        )
        result = 0
        for ring_set in ring_map:
            n_conj = sum(conj_rings_map[rnum] for rnum in ring_set)
            result += n_conj == 2
        return result

    def n_ring_system_with_conj_non_conj_simple_ring(self, mol: dm.Mol):
        """39 number of ring system containing 2 simple rings, one conjugated and one nonconjugated"""
        simple_rings, _, ring_map = _get_ring_system(mol)
        conj_rings_map = dict(
            (i, _is_ring_fully_conjugated(mol, x)) for i, x in enumerate(simple_rings)
        )
        result = 0
        for ring_set in ring_map:
            if len(ring_set) == 2:
                n_conj = sum(conj_rings_map[rnum] for rnum in ring_set)
                result += n_conj == 1
        return result

    def n_ring_system_with_3_conj_simple_ring(self, mol: dm.Mol):
        """40. number of rings systems with 3 conjugated simple rings"""
        simple_rings, _, ring_map = _get_ring_system(mol)
        conj_rings_map = dict(
            (i, _is_ring_fully_conjugated(mol, x)) for i, x in enumerate(simple_rings)
        )
        result = 0
        for ring_set in ring_map:
            n_conj = sum(conj_rings_map[rnum] for rnum in ring_set)
            result += n_conj == 3
        return result

    def n_ring_system_with_3_non_conj_simple_ring(self, mol: dm.Mol):
        """41. number of rings systems with 3 non-conjugated simple rings"""
        simple_rings, _, ring_map = _get_ring_system(mol)
        conj_rings_map = dict(
            (i, _is_ring_fully_conjugated(mol, x)) for i, x in enumerate(simple_rings)
        )
        result = 0
        for ring_set in ring_map:
            n_not_conj = sum(not conj_rings_map[rnum] for rnum in ring_set)
            result += n_not_conj == 3
        return result

    def n_ring_system_with_greater_one_conj_nonconj_simple_ring(self, mol: dm.Mol):
        """42. number of ring system containing 3 simple rings, at least one conjugated and one nonconjugated"""
        simple_rings, _, ring_map = _get_ring_system(mol)
        conj_rings_map = dict(
            (i, _is_ring_fully_conjugated(mol, x)) for i, x in enumerate(simple_rings)
        )
        result = 0
        for ring_set in ring_map:
            if len(ring_set) == 3:
                n_conj = sum(conj_rings_map[rnum] for rnum in ring_set)
                result += n_conj in [1, 2]
        return result

    @property
    def columns(self):
        """Get the name of all the descriptors of this calculator"""
        return list(self.DESCRIPTORS)

    def __call__(self, mol: Union[dm.Mol, str]):
        r"""
        Compute the Fingerprint of a molecule

        Args:
            mol: the molecule of interest

        Returns:
            props (np.ndarray): list of computed rdkit molecular descriptors
        """
        mol = dm.to_mol(mol)
        if self.use_scaffold and mol is not None:
            mol = MurckoScaffold.GetScaffoldForMol(mol)

        props = []
        for k in self.DESCRIPTORS:
            try:
                fn = getattr(self, k)
                props.append(fn(mol))
            except Exception as e:
                if self.verbose:
                    logger.error(e)
                props.append(float("nan"))
        props = np.asarray(props)
        if self.normalize:
            return self.compute_normalization(props)
        return props


def skdistance(
    sk1: np.ndarray,
    sk2: np.ndarray,
    weights: Optional[np.ndarray] = None,
    cdist: bool = False,
):
    """Compute the scaffold distance between two scaffold keys
    as described in https://pubs.acs.org/doi/abs/10.1021/ci5001983.
    The input features are expected to be normalized beforehand (see paper)

    Args:
        sk1: scaffold key 1
        sk2: scaffold key 2
        weights: how to weight each of the features. By default rank ordering is used.
        cdist: whether to compute the features on a batched of inputs (expected 2D)

    Returns:
        dist (float): distance between two scaffold keys
    """
    if weights is None:
        weights = 1 / (np.arange(sk1.shape[-1]) + 1)

    if cdist:
        sk1 = np.atleast_2d(sk1)
        sk2 = np.atleast_2d(sk2)
        val = np.abs(sk1[:, None] - sk2[:]) ** 1.5
        dist = np.sum(val * weights, axis=-1)
    else:
        if any((sk.ndim > 1 and sk.shape[0] != 1) for sk in [sk1, sk2]):
            raise ValueError("`cdist` mode was not detected, you need to provide single vectors")
        val = np.abs(sk1 - sk2) ** 1.5
        dist = np.sum(val * weights)
    return dist



from .base import SerializableCalculator, _CALCULATORS
from .cats import CATS
from .descriptors import RDKitDescriptors2D, RDKitDescriptors3D, MordredDescriptors
from .fingerprints import FPCalculator, FP_FUNCS
from .pharmacophore import Pharmacophore2D, Pharmacophore3D
from .shape import ElectroShapeDescriptors, USRDescriptors
from .skeys import ScaffoldKeyCalculator


def get_calculator(name: str, **params):
    """Get molecular calculator based on name

    Args:
        name: Name of the featurizer
        params (dict): Parameters of the featurizer

    Raises:
        ValueError: When featurizer is not supported

    Returns:
        featurizer: Callable
    """
    if not isinstance(name, str):
        return name

    CALC_MAP = {k.lower(): v for k, v in _CALCULATORS.items()}
    name = name.lower()
    if name in FP_FUNCS.keys():
        featurizer = FPCalculator(name, **params)
    elif name == "desc3d":
        featurizer = RDKitDescriptors3D(**params)
    elif name == "desc2d":
        featurizer = RDKitDescriptors2D(**params)
    elif name == "mordred":
        featurizer = MordredDescriptors(**params)
    elif name == "cats":
        featurizer = CATS(**params)
    elif name == "cats2d":
        params["use_3d_distances"] = False
        featurizer = CATS(**params)
    elif name == "cats3d":
        params["use_3d_distances"] = True
        featurizer = CATS(**params)
    elif name == "pharm2d":
        featurizer = Pharmacophore2D(**params)
    elif name == "pharm3d":
        featurizer = Pharmacophore3D(**params)
    elif name.startswith("usr"):
        params["method"] = name
        featurizer = USRDescriptors(**params)
    elif name == "electroshape":
        featurizer = ElectroShapeDescriptors(**params)
    elif name in ["scaffoldkeys", "skeys", "scaffkeys"]:
        featurizer = ScaffoldKeyCalculator(**params)
    elif name == "none":
        featurizer = None
    # for any generic calculator that has been automatically registered
    elif name in CALC_MAP.keys():
        featurizer = CALC_MAP[name](**params)
    else:
        raise ValueError(f"{name} is not a supported internal featurizer")
    return featurizer



import datamol as dm
import numpy as np
import pytest

from molfeat._version import __version__ as MOLFEAT_VERSION
from molfeat.calc import (
    CATS,
    FPCalculator,
    Pharmacophore2D,
    RDKitDescriptors2D,
    ScaffoldKeyCalculator,
)
from molfeat.calc._atom_bond_features import atom_chiral_tag_one_hot, atom_one_hot
from molfeat.calc.atom import AtomCalculator, AtomMaterialCalculator
from molfeat.calc.bond import BondCalculator
from molfeat.trans.base import MoleculeTransformer, PrecomputedMolTransformer
from molfeat.trans.fp import FPVecFilteredTransformer, FPVecTransformer
from molfeat.trans.graph import (
    AdjGraphTransformer,
    DGLGraphTransformer,
    MolTreeDecompositionTransformer,
    PYGGraphTransformer,
    TopoDistGraphTransformer,
)
from molfeat.trans.pretrained import (
    GraphormerTransformer,
    PretrainedDGLTransformer,
    PretrainedHFTransformer,
)
from molfeat.utils.cache import FileCache, MolToKey
from molfeat.utils.state import compare_state


def _dummy_featurizer_fn(x):
    return np.random.random(1024)


FEATURIZERS_SPEC = {
    # fp
    "FPVecTransformer": lambda: FPVecTransformer(
        kind="ecfp:4",
        length=2044,
        useFeatures=True,
        dtype=np.float64,
    ),
    "MoleculeTransformerFP": lambda: MoleculeTransformer(
        FPCalculator("ecfp", radius=2),
        dtype=np.float64,
    ),
    "FPVecFilteredTransformer": lambda: FPVecFilteredTransformer(),
    # base
    "MoleculeTransformer": lambda: MoleculeTransformer(featurizer=_dummy_featurizer_fn),
    # pretrained
    "PretrainedDGLTransformer": lambda: PretrainedDGLTransformer(),
    "GraphormerTransformer": lambda: GraphormerTransformer(),
    "PretrainedHFTransformer": lambda: PretrainedHFTransformer(),
    # graph
    "AdjGraphTransformer_with_bonds": lambda: AdjGraphTransformer(
        atom_featurizer=AtomMaterialCalculator(),
        bond_featurizer=BondCalculator(),
    ),
    "AdjGraphTransformer_with_bonds_custom_atom": lambda: AdjGraphTransformer(
        atom_featurizer=AtomMaterialCalculator(
            featurizer_funcs={
                "atom_chiral_tag_one_hot": atom_chiral_tag_one_hot,
                "atom_one_hot": atom_one_hot,
            }
        ),
        bond_featurizer=BondCalculator(),
    ),
    "AdjGraphTransformer": lambda: AdjGraphTransformer(atom_featurizer=AtomCalculator()),
    "DGLGraphTransformer": lambda: DGLGraphTransformer(),
    "TopoDistGraphTransformer": lambda: TopoDistGraphTransformer(),
    "PYGGraphTransformer": lambda: PYGGraphTransformer(),
    "MolTreeDecompositionTransformer": lambda: MolTreeDecompositionTransformer(verbose=False),
}
FEATURIZERS_ATOM_PICKLES = ["AdjGraphTransformer_with_bonds_custom_atom"]
FEATURIZERS_BUILDER_ATOM_PICKLES = [FEATURIZERS_SPEC[k] for k in FEATURIZERS_ATOM_PICKLES]
FEATURIZERS_NAMES, FEATURIZERS_BUILDER = zip(
    *[(k, v) for k, v in FEATURIZERS_SPEC.items() if k not in FEATURIZERS_ATOM_PICKLES]
)


@pytest.mark.parametrize(
    "featurizer_builder",
    FEATURIZERS_BUILDER,
    ids=FEATURIZERS_NAMES,
)
def test_to_from_state(featurizer_builder):
    featurizer: MoleculeTransformer = featurizer_builder()

    # check to_state
    state = featurizer.to_state_dict()
    assert "name" in state
    assert "args" in state
    assert "_molfeat_version" in state

    # check from_state
    featurizer = MoleculeTransformer.from_state_dict(state)
    state2 = featurizer.to_state_dict()
    assert state == state2


@pytest.mark.parametrize(
    "featurizer_builder",
    FEATURIZERS_BUILDER,
    ids=FEATURIZERS_NAMES,
)
def test_to_from_state_yaml(featurizer_builder, tmp_path):
    featurizer: MoleculeTransformer = featurizer_builder()

    featurizer_path = tmp_path / "featurizer.yaml"

    featurizer.to_state_yaml_file(featurizer_path)
    state = featurizer.to_state_dict()

    featurizer = MoleculeTransformer.from_state_yaml_file(featurizer_path)
    state2 = featurizer.to_state_dict()
    assert state == state2


@pytest.mark.parametrize(
    "featurizer_builder",
    FEATURIZERS_BUILDER,
    ids=FEATURIZERS_NAMES,
)
def test_to_from_state_json(featurizer_builder, tmp_path):
    featurizer: MoleculeTransformer = featurizer_builder()

    featurizer_path = tmp_path / "featurizer.json"

    featurizer.to_state_json_file(featurizer_path)
    state = featurizer.to_state_dict()

    featurizer = MoleculeTransformer.from_state_json_file(featurizer_path)
    state2 = featurizer.to_state_dict()
    assert state == state2


@pytest.mark.parametrize(
    "featurizer_builder",
    FEATURIZERS_BUILDER_ATOM_PICKLES,
    ids=FEATURIZERS_ATOM_PICKLES,
)
def test_state_atom_bond_pickle(featurizer_builder, tmp_path):
    featurizer: MoleculeTransformer = featurizer_builder()

    # check to_state
    state = featurizer.to_state_dict()
    assert "name" in state
    assert "args" in state
    assert "_molfeat_version" in state

    # check from_state
    featurizer = MoleculeTransformer.from_state_dict(state)
    state2 = featurizer.to_state_dict()
    featurizer2 = MoleculeTransformer.from_state_dict(state2)
    to_remove = []
    for key in state["args"]:
        if key.endswith("is_pickled") and state["args"][key] is True:
            to_remove.append(key.replace("is_pickled", "").strip("_"))
    for key_val in to_remove:
        state["args"].pop(key_val, None)
        state2["args"].pop(key_val, None)
    assert state == state2
    mols = dm.freesolv().smiles.iloc[:10]
    out = featurizer(mols)
    out2 = featurizer2(mols)
    assert len(out) == len(out2) == len(mols)
    for o1, o2 in zip(out, out2):
        if isinstance(o1, (list, tuple)):
            assert bool(np.all([np.allclose(i, j, atol=2) for i, j in zip(o1, o2)])) is True
        else:
            np.testing.assert_array_equal(o1, o2)


def test_PrecomputedMolTransformer_state(tmp_path):
    cache_path = tmp_path / "cache.parquet"
    smiles_list = ["[NH3+]CCSc1cc(-c2ccc[nH]2)c2c3c(ccc(F)c13)NC2=O", "CCCO"]

    featurizer = FPVecTransformer(kind="ecfp", length=2048)
    cache = FileCache(cache_file=cache_path, file_type="parquet", n_jobs=-1)

    featurizer_cache = PrecomputedMolTransformer(cache=cache, featurizer=featurizer)
    featurizer_cache(smiles_list)

    # sanity check
    assert len(featurizer_cache.cache.cache) == 2

    # serialize to state
    state = featurizer_cache.to_state_dict()

    assert state["name"] == "PrecomputedMolTransformer"
    assert "base_featurizer" in state
    assert "cache" in state
    assert "_molfeat_version" in state

    # reload from state
    featurizer_cache2: PrecomputedMolTransformer = MoleculeTransformer.from_state_dict(state)  # type: ignore

    assert featurizer_cache2.base_featurizer.to_state_dict() == featurizer.to_state_dict()
    assert len(featurizer_cache2.cache.cache) == 2
    assert all(
        featurizer_cache2.cache.to_dataframe(pack_bits=True)
        == featurizer_cache2.cache.to_dataframe(pack_bits=True)
    )


def test_PrecomputedMolTransformer_init_from_state_file(tmp_path):
    cache_path = str(tmp_path / "cache.parquet")
    smiles_list = ["[NH3+]CCSc1cc(-c2ccc[nH]2)c2c3c(ccc(F)c13)NC2=O", "CCCO"]

    featurizer = FPVecTransformer(kind="ecfp", length=2048)
    cache = FileCache(cache_file=cache_path, file_type="parquet", n_jobs=-1)

    featurizer_cache = PrecomputedMolTransformer(cache=cache, featurizer=featurizer)
    featurizer_cache(smiles_list)

    # sanity check
    assert len(featurizer_cache.cache.cache) == 2

    # serialize to state
    state = featurizer_cache.to_state_dict()

    # export yaml state file
    state_path = str(tmp_path / "state.yaml")

    featurizer_cache.to_state_yaml_file(filepath=state_path)

    assert state["name"] == "PrecomputedMolTransformer"
    assert "base_featurizer" in state
    assert "cache" in state
    assert "_molfeat_version" in state

    # reload from state
    featurizer_cache2: PrecomputedMolTransformer = PrecomputedMolTransformer(state_path=state_path)

    assert featurizer_cache2.base_featurizer.to_state_dict() == featurizer.to_state_dict()
    assert len(featurizer_cache2.cache.cache) == 2
    assert all(
        featurizer_cache2.cache.to_dataframe(pack_bits=True)
        == featurizer_cache2.cache.to_dataframe(pack_bits=True)
    )


def test_moltokey_state():
    hasher = MolToKey()
    state = hasher.to_state_dict()

    assert state == {"hash_name": "dm.unique_id"}

    hasher = MolToKey.from_state_dict(state)
    assert hasher.hash_name == "dm.unique_id"
    assert hasher.hash_fn == dm.unique_id


def test_fp_state():
    expected_res = [
        {
            "name": "FPCalculator",
            "module": "molfeat.calc.fingerprints",
            "args": {"length": 512, "method": "ecfp", "counting": False, "nBits": 512},
            "_molfeat_version": MOLFEAT_VERSION,
        },
        {
            "name": "FPCalculator",
            "module": "molfeat.calc.fingerprints",
            "args": {
                "length": 241,
                "method": "fcfp-count",
                "counting": True,
                "nBits": 241,
            },
            "_molfeat_version": MOLFEAT_VERSION,
        },
        {
            "name": "FPCalculator",
            "module": "molfeat.calc.fingerprints",
            "args": {"length": 2048, "method": "maccs", "counting": False},
            "_molfeat_version": MOLFEAT_VERSION,
        },
    ]
    for i, (fp_name, fp_len) in enumerate(zip(["ecfp", "fcfp-count", "maccs"], [512, 241, 2048])):
        calc = FPCalculator(fp_name, fp_len)
        state = calc.to_state_dict()
        assert state == expected_res[i]
    # check reloading
    calc = FPCalculator("ecfp", length=512, radius=3, useChirality=True)
    state = calc.to_state_dict()
    mol = dm.to_mol("C1CC(=O)NC(=O)[C@H]1N2C(=O)C3=CC=CC=C3C2=O")  # isomeric thalidomide
    expected = calc(mol)
    reloaded_calc = FPCalculator.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_rdkit_descr_state():
    calc = RDKitDescriptors2D(replace_nan=True, augment=False)
    expected_state = {
        "name": "RDKitDescriptors2D",
        "module": "molfeat.calc.descriptors",
        "args": {
            "replace_nan": True,
            "augment": False,
            "descrs": None,
            "avg_ipc": True,
            "do_not_standardize": False,
        },
        "_molfeat_version": MOLFEAT_VERSION,
    }

    state = calc.to_state_dict()
    assert state == expected_state
    mol = dm.to_mol("Nc1cnn(-c2ccccc2)c(=O)c1Cl")
    expected = calc(mol)
    reloaded_calc = RDKitDescriptors2D.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_pharmacophore_state():
    calc = Pharmacophore2D(factory="gobbi", length=1024, useCounts=False)
    expected_state = {
        "name": "Pharmacophore2D",
        "module": "molfeat.calc.pharmacophore",
        "args": {
            "factory": "gobbi",
            "useCounts": False,
            "minPointCount": 2,
            "maxPointCount": 3,
            "shortestPathsOnly": True,
            "includeBondOrder": False,
            "skipFeats": [],
            "trianglePruneBins": True,
            "bins": [(2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 100)],
            "length": 1024,
        },
        "_molfeat_version": MOLFEAT_VERSION,
    }

    state = calc.to_state_dict()
    assert state == expected_state
    mol = dm.to_mol("Nc1cnn(-c2ccccc2)c(=O)c1Cl")
    expected = calc(mol)
    reloaded_calc = Pharmacophore2D.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_cats_state():
    calc = CATS(max_dist=8, scale="count", bins=[0, 3, 6, 8])
    expected_state = {
        "name": "CATS",
        "module": "molfeat.calc.cats",
        "args": {
            "max_dist": 8,
            "bins": [0, 3, 6, 8],
            "scale": "count",
            "use_3d_distances": False,
        },
        "_molfeat_version": MOLFEAT_VERSION,
    }

    state = calc.to_state_dict()
    assert state == expected_state
    mol = dm.to_mol("Nc1cnn(-c2ccccc2)c(=O)c1Cl")
    expected = calc(mol)
    reloaded_calc = CATS.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_skeys_state():
    calc = ScaffoldKeyCalculator(normalize=True, use_scaffold=False, verbose=False)
    expected_state = {
        "name": "ScaffoldKeyCalculator",
        "module": "molfeat.calc.skeys",
        "args": {
            "normalize": True,
            "verbose": False,
            "use_scaffold": False,
        },
        "_molfeat_version": MOLFEAT_VERSION,
    }

    state = calc.to_state_dict()
    assert state == expected_state
    mol = dm.to_mol("Nc1cnn(-c2ccccc2)c(=O)c1Cl")
    expected = calc(mol)
    reloaded_calc = ScaffoldKeyCalculator.from_state_dict(state)
    out = reloaded_calc(mol)
    np.testing.assert_array_almost_equal(expected, out)


def test_filecache_state(tmp_path):
    # setup
    cache_path = tmp_path / "cache.parquet"
    cache = FileCache(cache_file=cache_path, file_type="parquet")
    cache(["CCC", "CCCCO"], featurizer=lambda x: [[0, 1, 1, 0], [0, 1, 1, 1]])

    # save state
    state = cache.to_state_dict(save_to_file=True)

    assert set(state.keys()) == {
        "_cache_name",
        "cache_file",
        "file_type",
        "mol_hasher",
        "n_jobs",
        "name",
        "clear_on_exit",
        "parquet_kwargs",
        "verbose",
    }

    assert dm.fs.exists(cache_path)

    # reload from state
    cache2 = FileCache.from_state_dict(state)
    assert len(cache2.cache) == 2
    assert all(cache.to_dataframe(pack_bits=True) == cache2.to_dataframe(pack_bits=True))


def test_state_comparison():
    state_dict1 = dict(_molfeat_version="0.1.2", value=2)
    state_dict2 = dict(_molfeat_version="0.1.2dev", value=2)
    state_dict3 = dict(_molfeat_version="0.1", value=2)
    state_dict4 = dict(value=2)
    state_dict5 = dict(_molfeat_version="1.2.3", value=100)
    state_dict6 = dict(_molfeat_version="0.1.2pre", value=0)

    expected_results = {
        "major": [True, True, True, False, False, False],
        1: [True, True, True, False, False, False],
        "micro": [True, True, False, False, False, False],
        None: [True, False, False, False, False, False],
    }
    for allowed_version in ["major", 1, "micro", None]:
        results = []
        results.append(compare_state(state_dict1, state_dict1, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict2, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict3, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict4, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict5, allow_version_level=allowed_version))
        results.append(compare_state(state_dict1, state_dict6, allow_version_level=allowed_version))
        assert expected_results[allowed_version] == results



import types
import datamol as dm


class SmilesConverter:
    """Molecule line notation conversion from smiles to selfies or inchi"""

    SUPPORTED_LINE_NOTATIONS = [
        "none",
        "smiles",
        "selfies",
        "inchi",
    ]

    def __init__(self, target: str = None):
        """
        Convert input smiles to a target line notation

        Args:
            target: target representation.
        """
        self.target = target

        if self.target is not None and self.target not in self.SUPPORTED_LINE_NOTATIONS:
            raise ValueError(
                f"{target} is not a supported line representation. Choose from {self.SUPPORTED_LINE_NOTATIONS}"
            )

        if self.target == "smiles" or (self.target is None or self.target == "none"):
            self.converter = None
        elif self.target == "inchi":
            self.converter = types.SimpleNamespace(decode=dm.from_inchi, encode=dm.to_inchi)
        elif self.target == "selfies":
            self.converter = types.SimpleNamespace(decode=dm.from_selfies, encode=dm.to_selfies)

    def decode(self, inp: str):
        """Decode inputs into smiles

        Args:
            inp: input representation to decode
        """
        if self.converter is None:
            return inp
        with dm.without_rdkit_log():
            try:
                decoded = self.converter.decode(inp)
                return decoded.strip()
            except Exception:  # (deepsmiles.DecodeError, ValueError, AttributeError, IndexError):
                return None

    def encode(self, smiles: str):
        """Encode a input smiles into target line notation

        Args:
            smiles: input smiles to encode
        """
        if self.converter is None:
            return smiles
        with dm.without_rdkit_log():
            try:
                encoded = self.converter.encode(smiles)
                return encoded.strip()
            except Exception:
                return None



from typing import Optional
from typing import Union
import torch
import numpy as np
from packaging import version
from molfeat.calc.atom import AtomCalculator
from molfeat.calc.atom import AtomMaterialCalculator
from molfeat.calc.atom import DGLWeaveAtomCalculator
from molfeat.calc.atom import DGLCanonicalAtomCalculator
from molfeat.calc.bond import BondCalculator
from molfeat.calc.bond import EdgeMatCalculator
from molfeat.calc.bond import DGLCanonicalBondCalculator
from molfeat.calc.bond import DGLWeaveEdgeCalculator


DTYPES_MAPPING = {
    None: None,
    torch.float16: "torch.float16",
    torch.float32: "torch.float32",
    torch.float64: "torch.float64",
    torch.int8: "torch.int8",
    torch.int16: "torch.int16",
    torch.int32: "torch.int32",
    torch.int64: "torch.int64",
    torch.uint8: "torch.uint8",
    torch.bool: "torch.bool",
    np.float16: "np.float16",
    np.float32: "np.float32",
    np.float64: "np.float64",
    np.int8: "np.int8",
    np.int16: "np.int16",
    np.int32: "np.int32",
    np.int64: "np.int64",
    np.uint8: "np.uint8",
    np.uint16: "np.uint16",
    np.uint32: "np.uint32",
    np.uint64: "np.uint64",
    bool: "bool",
    int: "int",
    float: "float",
}

DTYPES_MAPPING_REVERSE = {v: k for k, v in DTYPES_MAPPING.items()}

ATOM_FEATURIZER_MAPPING = {
    AtomCalculator: "AtomCalculator",
    AtomMaterialCalculator: "AtomMaterialCalculator",
    DGLCanonicalAtomCalculator: "DGLCanonicalAtomCalculator",
    DGLWeaveAtomCalculator: "DGLWeaveAtomCalculator",
}

ATOM_FEATURIZER_MAPPING_REVERSE = {v: k for k, v in ATOM_FEATURIZER_MAPPING.items()}

BOND_FEATURIZER_MAPPING = {
    BondCalculator: "BondCalculator",
    EdgeMatCalculator: "EdgeMatCalculator",
    DGLCanonicalBondCalculator: "DGLCanonicalBondCalculator",
    DGLWeaveEdgeCalculator: "DGLWeaveEdgeCalculator",
}

BOND_FEATURIZER_MAPPING_REVERSE = {v: k for k, v in BOND_FEATURIZER_MAPPING.items()}


def map_dtype(dtype: Optional[Union[str, torch.dtype, np.dtype]]):
    """Map a dtype to a string representation or the other way around"""

    if isinstance(dtype, str):
        mapping = DTYPES_MAPPING_REVERSE
    else:
        mapping = DTYPES_MAPPING

    if dtype not in mapping:
        msg = f"{dtype} is not a valid dtype. The valid dtypes are {list(mapping.keys())}."
        raise ValueError(msg)

    return mapping[dtype]


def compare_state(
    state_dict_1, state_dict_2, allow_version_level: Optional[Union[str, int]] = None
):
    """Compare two state dict and allow version matching
    Args:
        state_dict_1: state dict of the first object
        state_dict_2: state dict of the second object
        allow_version_level: version level for comparison.
            One of 'major', 'minor', 'micro' or 0, 1, 2
    """
    state_dict_1 = state_dict_1.copy()
    state_dict_2 = state_dict_2.copy()
    version1 = state_dict_1.pop("_molfeat_version", "")
    version2 = state_dict_2.pop("_molfeat_version", "")
    try:
        version1 = version.parse(version1)
        release1 = version1.release
    except version.InvalidVersion:
        version1 = None
        release1 = None
    try:
        version2 = version.parse(version2)
        release2 = version2.release
    except version.InvalidVersion:
        version2 = None
        release2 = None

    version_names = ["major", "minor", "micro"]
    if isinstance(allow_version_level, str):
        allow_version_level = version_names.index(allow_version_level)
        # we want an exception if fail here

    if allow_version_level is None:
        version_comp = version1 == version2
        return (state_dict_1 == state_dict_2) and version_comp

    if release1 is None or release2 is None:
        version_comp = release1 == release2
    # pad both to the same length
    else:
        longest = len(release2) - len(release1)
        release1 += (0,) * max(0, longest)
        release2 += (0,) * max(0, -longest)
        version_comp = release1[: allow_version_level + 1] == release2[: allow_version_level + 1]
    return (state_dict_1 == state_dict_2) and version_comp



from typing import List
from typing import Union
from typing import Optional
from collections import defaultdict

import os
import pathlib
import platformdirs
import fsspec
import datamol as dm

from loguru import logger
from joblib import Memory
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import minimum_spanning_tree
from rdkit.Chem import GetSymmSSSR
from rdkit.Chem import MolFragmentToSmiles

MAX_W = 1000
MST_MAX_WEIGHT = 100


class TreeDecomposer:
    """
    Molecule to tree decomposition.
    The decomposition follows algorithm 2 of https://arxiv.org/pdf/1802.04364.pdf
    """

    def __init__(self, tmp_dir: os.PathLike = None, cache: bool = True):
        """Tree Decomposer with cache system

        Args:
            tmp_dir (os.PathLike, optional): cache path. Defaults to None.
            cache (bool, optional): Whether to cache the computation. Defaults to True.
        """
        self.cache = cache
        if tmp_dir is None and cache:
            cache_dir = pathlib.Path(platformdirs.user_cache_dir(appname="molfeat"))
            tmp_dir = cache_dir / "treedecomp"
        if tmp_dir:
            self.tmp_dir = pathlib.Path(str(tmp_dir))
            self.tmp_dir.mkdir(exist_ok=True, parents=True)
        else:
            self.tmp_dir = None
        memory = Memory(self.tmp_dir, verbose=1)
        if self.cache:
            self.decomposition_into_tree = memory.cache(
                self.decomposition_into_tree, ignore=["self", "mol"]
            )

    def decomposition_into_tree(self, mol: dm.Mol, inchikey: str = None):
        """
        Find the maximum spanning tree over all the clusters of a molecule

        Args:
            mol: The molecule of interest
            inchikey: Optional inchi key of the molecule

        Returns:
            (nodes, edges [, frags]): A tuple
            - nodes (List(List[int])): list of all the cluster formed by the decomposition Each cluster is a list of integers
                which represent the ids of the atoms in clusters.
            - edges (List(Tuple[int, int])): list of the edges.
                Each edges is a tuple of source node and destination node represented by their position in the list of clusters
            - frags (List[str]): list of strings representing the smiles of the clusters.
        """
        n_atoms = mol.GetNumAtoms()
        if n_atoms == 1:  # special case
            return [[0]], [], [dm.to_smiles(mol)]

        cliques = []
        for bond in mol.GetBonds():
            a1 = bond.GetBeginAtom().GetIdx()
            a2 = bond.GetEndAtom().GetIdx()
            if not bond.IsInRing():
                cliques.append([a1, a2])

        ssr = [list(x) for x in GetSymmSSSR(mol)]
        cliques.extend(ssr)

        nei_list = [[] for i in range(n_atoms)]
        for i, atoms in enumerate(cliques):
            for atom in atoms:
                nei_list[atom].append(i)

        # Merge Rings with intersection > 2 atoms
        for i in range(len(cliques)):
            if len(cliques[i]) <= 2:
                continue
            for atom in cliques[i]:
                for j in nei_list[atom]:
                    if i >= j or len(cliques[j]) <= 2:
                        continue
                    inter = set(cliques[i]) & set(cliques[j])
                    if len(inter) > 2:
                        cliques[i].extend(cliques[j])
                        cliques[i] = list(set(cliques[i]))
                        cliques[j] = []

        cliques = [c for c in cliques if len(c) > 0]
        nei_list = [[] for i in range(n_atoms)]
        for i, atoms in enumerate(cliques):
            for atom in atoms:
                nei_list[atom].append(i)

        # Build edges and add singleton cliques
        edges = defaultdict(int)
        for atom in range(n_atoms):
            if len(nei_list[atom]) <= 1:
                continue
            cnei = nei_list[atom]
            bonds = [c for c in cnei if len(cliques[c]) == 2]
            rings = [c for c in cnei if len(cliques[c]) > 4]
            if len(bonds) > 2 or (len(bonds) == 2 and len(cnei) > 2):
                # In general, if len(cnei) >= 3, a singleton should be added, but 1 bond + 2 ring is currently not dealt with.
                cliques.append([atom])
                c2 = len(cliques) - 1
                for c1 in cnei:
                    edges[(c1, c2)] = 1
            elif len(rings) > 2:  # Multiple (n>2) complex rings
                cliques.append([atom])
                c2 = len(cliques) - 1
                for c1 in cnei:
                    edges[(c1, c2)] = MST_MAX_WEIGHT - 1
            else:
                for i in range(len(cnei)):
                    for j in range(i + 1, len(cnei)):
                        c1, c2 = cnei[i], cnei[j]
                        inter = set(cliques[c1]) & set(cliques[c2])
                        if edges[(c1, c2)] < len(inter):
                            edges[(c1, c2)] = len(inter)  # cnei[i] < cnei[j] by construction

        edges = [u + (MST_MAX_WEIGHT - v,) for u, v in edges.items()]
        frags = [MolFragmentToSmiles(mol, cluster) for cluster in cliques]

        if len(edges) == 0:
            return cliques, edges, frags

        # Compute Maximum Spanning Tree
        row, col, data = zip(*edges)
        n_clique = len(cliques)
        clique_graph = csr_matrix((data, (row, col)), shape=(n_clique, n_clique))
        junc_tree = minimum_spanning_tree(clique_graph)
        row, col = junc_tree.nonzero()
        edges = [(row[i], col[i]) for i in range(len(row))]
        return cliques, edges, frags

    def __call__(
        self,
        mol: Union[dm.Mol, str],
        as_smiles: bool = True,
    ):
        """
        Find the maximum spanning tree over all the clusters of a molecule

        Args:
            mol: the molecule of interest
            as_smiles: Whether to return molecular fragment as smiles or molecules
        """
        mol = dm.to_mol(mol)
        cliques, edges, frags = self.decomposition_into_tree(mol, dm.to_inchikey(mol))
        if not as_smiles:
            frags = [dm.to_mol(x) for x in frags]
        return cliques, edges, frags

    def get_vocab(
        self,
        mol_list: List[Union[str, dm.Mol]],
        output_file: Optional[os.PathLike] = None,
        log: bool = False,
    ):
        r"""
        Generate the list of all possible fragments given a set of molecules.
        This can be useful to build the vocabulary of fragments that can be found in a dataset of molecule
        before doing some learning.

        Args:
            mol_list (Iterable[dm.Mol]): A collection of molecules
            output_file: path to a file that will be used to store the generated set of fragments.
            log (bool, optional): Whether to print intermediate results to stdout

        Returns:
            res (List[str]): List of the smiles of all fragments found in the molecule collection
        """
        res = set()
        decomposer = TreeDecomposer(tmp_dir=self.tmp_dir, cache=True)
        output = dm.parallelized(decomposer, mol_list)
        for _, _, frag in output:
            res.update(frag)
        if log:
            logger.debug("A vocab of {} elements has been generated".format(len(res)))
        res = sorted(res, key=len)
        if output_file:
            with fsspec.open(output_file, "w") as outstream:
                outstream.write("\n".join(res))
        return res



from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import torch
import copy
from molfeat.utils.commons import _parse_to_evaluable_str
from molfeat.utils import requires
from molfeat.utils.pooler import Pooling
from molfeat.trans.pretrained.base import PretrainedMolTransformer

if requires.check("graphormer_pretrained"):
    from graphormer_pretrained.embeddings import GraphormerEmbeddingsExtractor
    from graphormer_pretrained.data.smiles_dataset import GraphormerInferenceDataset


class GraphormerTransformer(PretrainedMolTransformer):
    r"""
    Graphormer transformer from microsoft, pretrained on PCQM4Mv2 quantum chemistry dataset
    for the prediction of homo-lumo gap.

    Attributes:
        featurizer: Graphormer embedding object
        dtype: Data type. Use call instead
        pooling: Pooling method for Graphormer's embedding layer
    """

    def __init__(
        self,
        kind: str = "pcqm4mv2_graphormer_base",
        dtype: Callable = np.float32,
        pooling: str = "mean",
        max_length: Optional[int] = None,
        concat_layers: Union[List[int], int] = -1,
        ignore_padding: bool = True,
        version=None,
        **params,
    ):
        """
        Pretrained graphormer featurizer.

        !!! note
            The default behaviour of this feature extractor is to return the last hidden state of the encoder,
            averaged across all nodes (including the virtual node connected to all other nodes).

            For a different behaviour, please change the pooling method:
            * `graph` or `virtual`: use the virtual node embedding in the last layer to get the graph representation
            * `mean`, `max`, `sum`, etc or any other supported pooling of `molfeat.utils.pooler.Pooling`
                will take the operation defined by the pooling layer across all nodes of each graph

        Args:
            kind: name of the featurizer as available in the model store
            dtype: Data type to output
            pooling: type of pooling to use. One of ['graph', 'virtual', 'mean', 'max', 'sum']. The value "graph" corresponds to the virtual node representation
            max_length: Maximum length of the input sequence to consider. Please update this for large sequences
            concat_layers: Layer to concat to get the representation. By default the last hidden layer is returned.
            ignore_padding: Whether to ignore padding in the representation (default: True) to avoid effect of batching
            params: any other parameter to pass to PretrainedMolTransformer
        """

        super().__init__(dtype=dtype, pooling=pooling, **params)
        if not requires.check("graphormer_pretrained"):
            raise ValueError("`graphormer` is required to use this featurizer.")

        if concat_layers is None:
            concat_layers = -1
        if not isinstance(concat_layers, list):
            concat_layers = [concat_layers]
        self.concat_layers = concat_layers
        self.preload = True
        self.name = kind
        self._require_mols = False
        self.max_length = max_length
        self.ignore_padding = ignore_padding
        if isinstance(pooling, str):
            if pooling in Pooling.SUPPORTED_POOLING:
                pooling = Pooling(dim=1, name=pooling)
            else:
                pooling = None
        self.pooling = pooling
        self.featurizer = GraphormerEmbeddingsExtractor(
            pretrained_name=self.name, max_nodes=self.max_length, concat_layers=self.concat_layers
        )
        self.featurizer.config.max_nodes = self.max_length
        self.version = version

    def __repr__(self):
        return "{}(name={}, pooling={}, dtype={})".format(
            self.__class__.__name__,
            _parse_to_evaluable_str(self.name),
            _parse_to_evaluable_str(self.pooling.name),
            _parse_to_evaluable_str(self.dtype),
        )

    @staticmethod
    def list_available_models():
        """List available graphormer model to use"""
        return [
            "pcqm4mv1_graphormer_base",  # PCQM4Mv1
            "pcqm4mv2_graphormer_base",  # PCQM4Mv2
            "pcqm4mv1_graphormer_base_for_molhiv",  # ogbg-molhiv
            "oc20is2re_graphormer3d_base",  # Open Catalyst Challenge
        ]

    def _embed(self, inputs: List[str], **kwargs):
        """Internal molecular embedding

        Args:
            smiles: input smiles
        """
        with torch.no_grad():
            layer_reprs, graph_reprs, padding_mask = self.featurizer(inputs)
            if self.pooling is None:
                x = graph_reprs
            else:
                x = self.pooling(layer_reprs, mask=(padding_mask if self.ignore_padding else None))
        return x.numpy()

    def __getstate__(self):
        """Getting state to allow pickling"""
        d = copy.deepcopy(self.__dict__)
        d["precompute_cache"] = None
        d.pop("featurizer", None)
        return d

    def __setstate__(self, d):
        """Setting state during reloading pickling"""
        self.__dict__.update(d)
        self._update_params()

    def compute_max_length(self, inputs: list):
        """Compute maximum node number for the input list of molecules

        Args:
            inputs: input list of molecules
        """
        dataset = GraphormerInferenceDataset(
            inputs,
            multi_hop_max_dist=self.featurizer.config.multi_hop_max_dist,
            spatial_pos_max=self.featurizer.config.spatial_pos_max,
        )
        xs = [item.x.size(0) for item in dataset]
        return max(xs)

    def set_max_length(self, max_length: int):
        """Set the maximum length for this featurizer"""
        self.max_length = max_length
        self._update_params()
        self._preload()

    def _convert(self, inputs: list, **kwargs):
        """Convert molecules to the right format

        Args:
            inputs: inputs to preprocess

        Returns:
            processed: pre-processed input list
        """
        inputs = super()._convert(inputs, **kwargs)
        batch = self.featurizer._convert(inputs)
        return batch

    def _update_params(self):
        super()._update_params()
        self.featurizer = GraphormerEmbeddingsExtractor(
            pretrained_name=self.name, max_nodes=self.max_length
        )
        self.featurizer.config.max_nodes = self.max_length
